import dataclasses
from functools import partial
from typing import List, Optional

import pytest
from aioredis import RedisError

from dbdaora import (
    CacheType,
    DictFallbackDataSource,
    HashRepository,
    HashService,
    build_service,
    make_aioredis_data_source,
)


@pytest.mark.asyncio
@pytest.fixture
async def fake_service(mocker, fallback_data_source, fake_hash_repository_cls):
    memory_data_source_factory = partial(
        make_aioredis_data_source,
        'redis://',
        'redis://localhost/1',
        'redis://localhost/2',
    )

    async def fallback_data_source_factory():
        return fallback_data_source

    service = await build_service(
        HashService,
        fake_hash_repository_cls,
        memory_data_source_factory,
        fallback_data_source_factory,
        repository_expire_time=1,
        cache_type=CacheType.TTL,
        cache_ttl=1,
        cache_max_size=1,
        cb_failure_threshold=0,
        cb_recovery_timeout=10,
        cb_expected_exception=RedisError,
        logger=mocker.MagicMock(),
    )

    yield service

    service.repository.memory_data_source.close()
    await service.repository.memory_data_source.wait_closed()


@pytest.fixture
def fallback_data_source():
    return DictFallbackDataSource()


@dataclasses.dataclass
class FakeInnerEntity:
    id: str


@dataclasses.dataclass
class FakeEntity:
    id: str
    integer: int
    inner_entities: List[FakeInnerEntity]
    number: Optional[float] = None
    boolean: Optional[bool] = None


class FakeHashRepository(HashRepository[str]):
    name = 'fake'
    entity_cls = FakeEntity


@pytest.fixture
def fake_hash_repository_cls():
    return FakeHashRepository


@pytest.fixture
def dict_repository_cls():
    return FakeHashRepository


@pytest.fixture
def fake_entity():
    return FakeEntity(
        id='fake',
        inner_entities=[FakeInnerEntity('inner1'), FakeInnerEntity('inner2')],
        integer=1,
        number=0.1,
        boolean=True,
    )


@pytest.fixture
def fake_entity2():
    return FakeEntity(
        id='fake2',
        inner_entities=[FakeInnerEntity('inner3'), FakeInnerEntity('inner4')],
        integer=2,
        number=0.2,
        boolean=False,
    )


@pytest.fixture
def serialized_fake_entity():
    return {
        b'id': b'fake',
        b'integer': b'1',
        b'number': b'0.1',
        b'boolean': b'1',
        b'inner_entities': b'[{"id":"inner1"},{"id":"inner2"}]',
    }


@pytest.fixture
def serialized_fake_entity2():
    return {
        b'id': b'fake2',
        b'integer': b'2',
        b'number': b'0.2',
        b'boolean': b'0',
        b'inner_entities': b'[{"id":"inner3"},{"id":"inner4"}]',
    }


@pytest.mark.asyncio
@pytest.fixture
async def repository(mocker):
    memory_data_source = await make_aioredis_data_source(
        'redis://', 'redis://localhost/1', 'redis://localhost/2'
    )
    yield FakeHashRepository(
        memory_data_source=memory_data_source,
        fallback_data_source=DictFallbackDataSource(),
        expire_time=1,
    )
    memory_data_source.close()
    await memory_data_source.wait_closed()



from typing import Any, AsyncGenerator, Optional, Sequence

from cachetools import Cache

from ...keys import FallbackKey
from ...service import Service
from ..entity import SortedSetData, SortedSetEntity


class SortedSetService(Service[SortedSetEntity, SortedSetData, FallbackKey]):
    def get_many(
        self, *ids: str, **filters: Any
    ) -> AsyncGenerator[SortedSetEntity, None]:
        raise NotImplementedError()  # pragma: no cover

    def get_many_cached(
        self,
        ids: Sequence[str],
        cache: Cache,
        memory: bool = True,
        **filters: Any,
    ) -> AsyncGenerator[SortedSetEntity, None]:
        raise NotImplementedError()  # pragma: no cover

    async def delete(
        self, entity_id: Optional[str] = None, **filters: Any
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def exists(self, id: Optional[str] = None, **filters: Any) -> bool:
        raise NotImplementedError()  # pragma: no cover

    async def exists_cached(
        self, cache: Cache, memory: bool = True, **filters: Any,
    ) -> bool:
        raise NotImplementedError()  # pragma: no cover



import dataclasses
from functools import partial

import pytest
from aioredis import RedisError

from dbdaora import (
    BooleanRepository,
    BooleanService,
    CacheType,
    DictFallbackDataSource,
    build_service,
    make_aioredis_data_source,
)


@pytest.fixture
def memory_data_source_factory():
    return partial(
        make_aioredis_data_source,
        'redis://',
        'redis://localhost/1',
        'redis://localhost/2',
    )


@pytest.mark.asyncio
@pytest.fixture
async def fake_service(
    memory_data_source_factory,
    mocker,
    fallback_data_source,
    fake_boolean_repository_cls,
):
    async def fallback_data_source_factory():
        return fallback_data_source

    service = await build_service(
        BooleanService,
        fake_boolean_repository_cls,
        memory_data_source_factory,
        fallback_data_source_factory,
        repository_expire_time=1,
        cache_type=CacheType.TTL,
        cache_ttl=1,
        cache_max_size=2,
        cb_failure_threshold=0,
        cb_recovery_timeout=10,
        cb_expected_exception=RedisError,
        cb_expected_fallback_exception=KeyError,
        logger=mocker.MagicMock(),
    )

    yield service

    service.repository.memory_data_source.close()
    await service.repository.memory_data_source.wait_closed()


@pytest.fixture
def fallback_data_source():
    return DictFallbackDataSource()


@dataclasses.dataclass
class FakeEntity:
    id: str


class FakeBooleanRepository(BooleanRepository[str]):
    name = 'fake'
    id_name = 'id'
    entity_cls = FakeEntity


@pytest.fixture
def fake_entity_cls():
    return FakeEntity


@pytest.fixture
def fake_boolean_repository_cls():
    return FakeBooleanRepository


@pytest.fixture
def dict_repository_cls():
    return FakeBooleanRepository


@pytest.fixture
def fake_entity():
    return FakeEntity(id='fake',)


@pytest.fixture
def fake_entity2():
    return FakeEntity(id='fake2',)


@pytest.fixture
def serialized_fake_entity():
    return '1'


@pytest.fixture
def serialized_fake_entity2():
    return '1'


@pytest.mark.asyncio
@pytest.fixture
async def repository(
    mocker,
    memory_data_source_factory,
    fallback_data_source,
    fake_boolean_repository_cls,
):
    memory_data_source = await memory_data_source_factory()
    yield fake_boolean_repository_cls(
        memory_data_source=memory_data_source,
        fallback_data_source=fallback_data_source,
        expire_time=1,
    )
    memory_data_source.close()
    await memory_data_source.wait_closed()



from dataclasses import dataclass
from logging import Logger, getLogger
from typing import (
    Any,
    AsyncGenerator,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Union,
)

from cachetools import Cache

from dbdaora.exceptions import EntityNotFoundError, RequiredKeyAttributeError

from ..circuitbreaker import AsyncCircuitBreaker, DBDaoraCircuitBreakerError
from ..entity import Entity, EntityData
from ..keys import FallbackKey
from ..repository import MemoryRepository


@dataclass(init=False)
class Service(Generic[Entity, EntityData, FallbackKey]):
    repository: MemoryRepository[Entity, EntityData, FallbackKey]
    circuit_breaker: AsyncCircuitBreaker
    fallback_circuit_breaker: AsyncCircuitBreaker
    cache: Optional[Cache]
    logger: Logger

    def __init__(
        self,
        repository: MemoryRepository[Entity, EntityData, FallbackKey],
        circuit_breaker: AsyncCircuitBreaker,
        fallback_circuit_breaker: AsyncCircuitBreaker,
        cache: Optional[Cache] = None,
        exists_cache: Optional[Cache] = None,
        logger: Logger = getLogger(__name__),
    ):
        self.repository = repository
        self.circuit_breaker = circuit_breaker
        self.fallback_circuit_breaker = fallback_circuit_breaker
        self.cache = cache
        self.exists_cache = exists_cache
        self.entity_circuit = self.circuit_breaker(self.repository.entity)
        self.add_circuit = self.circuit_breaker(self.repository.add)
        self.delete_circuit = self.circuit_breaker(self.repository.delete)
        self.exists_circuit = self.circuit_breaker(self.repository.exists)
        self.logger = logger

        self.entity_fallback_circuit = self.fallback_circuit_breaker(
            self.repository.entity
        )
        self.add_fallback_circuit = self.fallback_circuit_breaker(
            self.repository.add
        )
        self.delete_fallback_circuit = self.fallback_circuit_breaker(
            self.repository.delete
        )
        self.exists_fallback_circuit = self.fallback_circuit_breaker(
            self.repository.exists
        )

    def get_many(
        self, *ids: str, memory: bool = True, **filters: Any
    ) -> AsyncGenerator[Entity, None]:
        if self.cache is None:
            if memory:
                return self.get_many_no_cache_memory(*ids, **filters)

            return self.get_many_no_cache_fallback(*ids, **filters)

        return self.get_many_cached(ids, self.cache, **filters)

    def should_raise_not_found_error_for_fallback_circuit_breaker(
        self, error: DBDaoraCircuitBreakerError
    ) -> bool:
        if self.fallback_circuit_breaker:
            exceptions = self.fallback_circuit_breaker.expected_exception

            if not isinstance(exceptions, tuple):
                exceptions = (exceptions,)

            for fallback_exception in exceptions:
                if isinstance(error, fallback_exception):
                    return True

        return False

    async def get_many_no_cache_memory(
        self, *ids: str, **filters: Any
    ) -> AsyncGenerator[Entity, None]:
        try:
            try:
                async for entity in self.repository.query(
                    many=ids, **filters
                ).entities:
                    yield entity

                self.circuit_breaker.set_success()

            except self.circuit_breaker.expected_exception as error:
                self.circuit_breaker.set_failure('get-many-memory', error)
                raise

        except DBDaoraCircuitBreakerError as error:
            self.logger.warning(error)
            if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                error
            ):
                raise EntityNotFoundError(ids, filters)

            try:
                try:
                    async for entity in self.repository.query(
                        many=ids, memory=False, **filters
                    ).entities:
                        yield entity

                    self.fallback_circuit_breaker.set_success()

                except self.fallback_circuit_breaker.expected_exception as error:
                    self.fallback_circuit_breaker.set_failure(
                        'get-many-fallback', error
                    )
                    raise

            except DBDaoraCircuitBreakerError as fallback_error:
                self.logger.warning(fallback_error)
                raise EntityNotFoundError(ids, filters)

    async def get_many_no_cache_fallback(
        self, *ids: str, **filters: Any
    ) -> AsyncGenerator[Entity, None]:
        try:
            try:
                async for entity in self.repository.query(
                    many=ids, memory=False, **filters
                ).entities:
                    yield entity

                self.fallback_circuit_breaker.set_success()

            except self.fallback_circuit_breaker.expected_exception as error:
                self.fallback_circuit_breaker.set_failure(
                    'get-many-fallback', error
                )
                raise

        except DBDaoraCircuitBreakerError as fallback_error:
            self.logger.warning(fallback_error)
            raise EntityNotFoundError(ids, filters)

    async def get_many_cached(
        self,
        ids: Sequence[str],
        cache: Cache,
        memory: bool = True,
        **filters: Any,
    ) -> AsyncGenerator[Entity, None]:
        missed_ids: List[str] = []
        missed_ids_set: Set[str] = set()
        found_ids_set: Set[str] = set()
        cache_key_suffix = self.cache_key_suffix(**filters)

        for id_ in ids:
            entity = self.get_cached_entity(id_, cache_key_suffix, **filters)

            if entity is None:
                missed_ids.append(id_)
                missed_ids_set.add(id_)

            elif entity is not CACHE_ALREADY_NOT_FOUND:
                yield entity

        if missed_ids:
            if memory:
                try:
                    try:
                        async for entity in self.repository.query(
                            many=missed_ids, **filters
                        ).entities:
                            yield entity
                            id_ = self.entity_id(entity)
                            self.set_cached_entity(
                                id_, cache_key_suffix, entity
                            )
                            found_ids_set.add(id_)

                        self.circuit_breaker.set_success()

                    except self.circuit_breaker.expected_exception as error:
                        self.circuit_breaker.set_failure(
                            'get-many-memory', error
                        )
                        raise

                except DBDaoraCircuitBreakerError as error:
                    self.logger.warning(error)
                    if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                        error
                    ):
                        raise EntityNotFoundError(ids, filters)

                    try:
                        try:
                            async for entity in self.repository.query(
                                many=missed_ids, memory=False, **filters
                            ).entities:
                                yield entity
                                id_ = self.entity_id(entity)
                                self.set_cached_entity(
                                    id_, cache_key_suffix, entity
                                )
                                found_ids_set.add(id_)

                            self.fallback_circuit_breaker.set_success()

                        except self.fallback_circuit_breaker.expected_exception as error:
                            self.fallback_circuit_breaker.set_failure(
                                'get-many-fallback', error
                            )
                            raise

                    except DBDaoraCircuitBreakerError as fallback_error:
                        self.logger.warning(fallback_error)
                        raise EntityNotFoundError(ids, filters)

            else:
                try:
                    try:
                        async for entity in self.repository.query(
                            many=missed_ids, memory=False, **filters
                        ).entities:
                            yield entity
                            id_ = self.entity_id(entity)
                            self.set_cached_entity(
                                id_, cache_key_suffix, entity
                            )
                            found_ids_set.add(id_)

                        self.fallback_circuit_breaker.set_success()

                    except self.fallback_circuit_breaker.expected_exception as error:
                        self.fallback_circuit_breaker.set_failure(
                            'get-many-fallback', error
                        )
                        raise
                except DBDaoraCircuitBreakerError as fallback_error:
                    self.logger.warning(fallback_error)
                    raise EntityNotFoundError(ids, filters)

        for id_ in missed_ids_set.difference(found_ids_set):
            self.set_cached_entity(
                id_, cache_key_suffix, CACHE_ALREADY_NOT_FOUND
            )

    def entity_id(self, entity: Entity) -> str:
        return (  # type: ignore
            entity[self.repository.id_name]
            if isinstance(entity, dict)
            else entity
            if isinstance(entity, str)
            else getattr(entity, self.repository.id_name)
        )

    def get_cached_entity(
        self, id: str, key_suffix: str, **filters: Any,
    ) -> Any:
        if self.cache is None:
            return None

        return self.cache.get(self.cache_key(id, key_suffix))

    def cache_key(self, id: str, suffix: str) -> str:
        return f'{id}{suffix}'

    def set_cached_entity(
        self,
        id: str,
        key_suffix: str,
        entity: Union[Entity, 'CacheAlreadyNotFound'],
    ) -> None:
        if self.cache is not None:
            self.cache[self.cache_key(id, key_suffix)] = entity

    def cache_key_suffix(self, **filters: Any) -> str:
        return (
            ''.join(f'{f}{v}' for f, v in filters.items()) if filters else ''
        )

    async def get_one(self, id: Optional[str] = None, **filters: Any) -> Any:
        if id is not None:
            filters['id'] = id

        try:
            if self.cache is None:
                return await self.entity_circuit(
                    self.repository.query(**filters)
                )

            return await self.get_one_cached(cache=self.cache, **filters)

        except DBDaoraCircuitBreakerError as error:
            self.logger.warning(error)
            if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                error
            ):
                raise EntityNotFoundError(id, filters)

            try:
                filters.pop('memory', None)

                if self.cache is not None:
                    return await self.get_one_cached(
                        cache=self.cache, memory=False, **filters
                    )

                return await self.entity_fallback_circuit(
                    self.repository.query(memory=False, **filters)
                )

            except DBDaoraCircuitBreakerError as fallback_error:
                self.logger.warning(fallback_error)
                raise EntityNotFoundError(id, filters)

    async def get_one_cached(
        self, cache: Cache, memory: bool = True, **filters: Any,
    ) -> Any:
        id = filters.pop(self.repository.id_name, None)

        if id is None:
            raise RequiredKeyAttributeError(
                type(self.repository).__name__,
                self.repository.id_name,
                self.repository.key_attrs,
            )

        cache_key_suffix = self.cache_key_suffix(**filters)
        entity = self.get_cached_entity(id, cache_key_suffix, **filters)

        if entity is None:
            filters[self.repository.id_name] = id

            try:
                if memory:
                    entity = await self.entity_circuit(
                        self.repository.query(**filters)
                    )
                else:
                    entity = await self.entity_fallback_circuit(
                        self.repository.query(memory=False, **filters)
                    )

                self.set_cached_entity(id, cache_key_suffix, entity)

            except EntityNotFoundError:
                self.set_cached_entity(
                    id, cache_key_suffix, CACHE_ALREADY_NOT_FOUND
                )
                raise

        elif entity == CACHE_ALREADY_NOT_FOUND:
            raise EntityNotFoundError(id)

        return entity

    async def add(
        self, entity: Any, *entities: Any, memory: bool = True
    ) -> None:
        if not memory:
            try:
                await self.add_fallback_circuit(
                    entity, memory=False, *entities
                )

            except DBDaoraCircuitBreakerError as fallback_error:
                self.logger.warning(fallback_error)
                raise

            return

        try:
            await self.add_circuit(entity, *entities)

        except DBDaoraCircuitBreakerError as error:
            self.logger.warning(error)
            if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                error
            ):
                raise

            try:
                await self.add_fallback_circuit(
                    entity, memory=False, *entities
                )
            except DBDaoraCircuitBreakerError as fallback_error:
                self.logger.warning(fallback_error)
                raise

    async def delete(
        self, entity_id: Optional[str] = None, **filters: Any
    ) -> None:
        if entity_id is not None:
            filters['id'] = entity_id

        try:
            await self.delete_circuit(self.repository.query(**filters))

        except DBDaoraCircuitBreakerError as error:
            self.logger.warning(error)
            if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                error
            ):
                raise

            try:
                await self.delete_fallback_circuit(
                    self.repository.query(memory=False, **filters)
                )

            except DBDaoraCircuitBreakerError as fallback_error:
                self.logger.warning(fallback_error)
                raise

    async def exists(self, id: Optional[str] = None, **filters: Any) -> bool:
        if id is not None:
            filters['id'] = id

        try:
            if self.exists_cache is None:
                return await self.exists_circuit(
                    self.repository.query(**filters)
                )

            return await self.exists_cached(self.exists_cache, **filters)

        except DBDaoraCircuitBreakerError as error:
            self.logger.warning(error)
            if self.should_raise_not_found_error_for_fallback_circuit_breaker(
                error
            ):
                raise

            if self.exists_cache is not None:
                return await self.exists_cached(
                    self.exists_cache, memory=False, **filters,
                )

            return await self.exists_fallback_circuit(
                self.repository.query(memory=False, **filters)
            )

    async def exists_cached(
        self, cache: Cache, memory: bool = True, **filters: Any,
    ) -> bool:
        id = filters.pop(self.repository.id_name, None)

        if id is None:
            raise RequiredKeyAttributeError(
                type(self.repository).__name__,
                self.repository.id_name,
                self.repository.key_attrs,
            )

        cache_key = self.cache_key(id, self.cache_key_suffix(**filters))
        entity_exists = cache.get(cache_key)

        if entity_exists is None:
            filters[self.repository.id_name] = id

            if memory:
                entity_exists = await self.exists_circuit(
                    self.repository.query(**filters)
                )
            else:
                entity_exists = await self.exists_fallback_circuit(
                    self.repository.query(memory=False, **filters)
                )

            if not entity_exists:
                cache[cache_key] = CACHE_ALREADY_NOT_FOUND
                return False
            else:
                cache[cache_key] = True

        elif entity_exists == CACHE_ALREADY_NOT_FOUND:
            return False

        return True

    async def shutdown(self) -> None:
        self.repository.memory_data_source.close()
        await self.repository.memory_data_source.wait_closed()


class CacheAlreadyNotFound:
    ...


CACHE_ALREADY_NOT_FOUND = CacheAlreadyNotFound()



import dataclasses
from typing import (
    Any,
    AsyncGenerator,
    ClassVar,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from dbdaora.entity import Entity, EntityData
from dbdaora.exceptions import RequiredKeyAttributeError
from dbdaora.keys import FallbackKey


@dataclasses.dataclass(init=False)
class BaseQuery(Generic[Entity, EntityData, FallbackKey]):
    repository: 'MemoryRepository[Entity, EntityData, FallbackKey]'
    memory: bool

    def __init__(
        self,
        repository: 'MemoryRepository[Entity, EntityData, FallbackKey]',
        *args: Any,
        memory: bool = True,
        **kwargs: Any,
    ):
        self.repository = repository
        self.memory = memory

    def make_key_parts(self, *args: Any, **kwargs: Any) -> List[Any]:
        missed_key_attrs = []
        key_parts = [
            args[i] if i < len(args) else missed_key_attrs.append((i, key_attr))  # type: ignore
            for i, key_attr in enumerate(self.repository.key_attrs)
        ]

        try:
            for i, attr_name in missed_key_attrs:
                key_parts[i] = kwargs[attr_name]
        except KeyError:
            raise RequiredKeyAttributeError(
                type(self.repository).__name__,
                attr_name,
                self.repository.key_attrs,
            )

        return key_parts

    @property
    async def entity(self) -> Entity:
        raise NotImplementedError()  # pragma: no cover

    @property
    def entities(self) -> AsyncGenerator[Entity, None]:
        raise NotImplementedError()  # pragma: no cover

    @property
    async def delete(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    @property
    async def exists(self) -> bool:
        raise NotImplementedError()  # pragma: no cover


class Query(BaseQuery[Entity, EntityData, FallbackKey]):
    key_parts: List[Any]

    def __init__(
        self,
        repository: 'MemoryRepository[Entity, EntityData, FallbackKey]',
        *args: Any,
        memory: bool = True,
        key_parts: Optional[List[Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(repository, memory=memory)

        if key_parts:
            self.key_parts = key_parts
        else:
            self.key_parts = self.make_key_parts(*args, **kwargs)

    @property
    async def delete(self) -> None:
        await self.repository.delete(self)

    @property
    async def add(self) -> None:
        await self.repository.add(query=self)

    @property
    async def entity(self) -> Entity:
        return await self.repository.entity(self)

    @property
    async def exists(self) -> bool:
        return await self.repository.exists(self)


class QueryMany(BaseQuery[Entity, EntityData, FallbackKey]):
    query_cls: ClassVar[Type[Query[Entity, EntityData, FallbackKey]]] = Query[
        Entity, EntityData, FallbackKey
    ]
    queries: List[Query[Entity, EntityData, FallbackKey]]

    def __init__(
        self,
        repository: 'MemoryRepository[Entity, EntityData, FallbackKey]',
        *args: Any,
        many: List[Union[Any, Tuple[Any, ...]]],
        memory: bool = True,
        many_key_parts: Optional[List[List[Any]]] = None,
        **kwargs: Any,
    ):
        self.memory = memory
        self.repository = repository

        if many_key_parts is None:
            many_key_parts = []
            many_key_attrs = (
                (repository.many_key_attrs,)
                if isinstance(repository.many_key_attrs, str)
                else repository.many_key_attrs
            )

            for many_input in many:
                kwargs_i = {}

                if isinstance(many_input, tuple):
                    start_index = len(many_key_attrs) - len(many_input)
                    start_index = start_index if start_index >= 0 else 0
                    kwargs_i.update(
                        {
                            name: input_
                            for name, input_ in zip(
                                many_key_attrs[start_index:], many_input
                            )
                        }
                    )
                else:
                    kwargs_i[many_key_attrs[-1]] = many_input

                many_key_parts.append(
                    self.make_key_parts(*args, **kwargs, **kwargs_i)
                )

        self.queries = [
            self.query_cls(repository, memory=memory, key_parts=key_parts)
            for key_parts in many_key_parts
        ]

    @property
    def entities(self) -> AsyncGenerator[Entity, None]:
        return self.repository.entities(self)


def make(
    *args: Any, **kwargs: Any
) -> BaseQuery[Entity, EntityData, FallbackKey]:
    if kwargs.get('many') or kwargs.get('many_key_parts'):
        return QueryMany(*args, **kwargs)

    return Query(*args, **kwargs)


from .repository import MemoryRepository  # noqa isort:skip



from hashlib import md5
from typing import Generic, Optional, Sequence, TypeVar


DataSource = TypeVar('DataSource')


class HashRing(Generic[DataSource]):
    def __init__(
        self, nodes: Sequence[DataSource], nodes_size: Optional[int] = None
    ):
        self.nodes = nodes

        if nodes_size is None:
            nodes_size = len(nodes)

        self.nodes_size = nodes_size

    def get_index(self, key: str) -> int:
        return (
            int(md5(str(key).encode('utf-8')).hexdigest(), 16)
            % self.nodes_size
        )

    def get_node(self, key: str) -> DataSource:
        return self.nodes[self.get_index(key)]



import dataclasses
import re
from typing import (  # type: ignore
    Any,
    AsyncGenerator,
    ClassVar,
    Generic,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    _TypedDictMeta,
)

from dbdaora import FallbackDataSource, MemoryDataSource
from dbdaora.entity import EntityData
from dbdaora.exceptions import (
    EntityNotFoundError,
    InvalidKeyAttributeError,
    InvalidQueryError,
    RequiredClassAttributeError,
)
from dbdaora.keys import FallbackKey

from .entity import Entity


@dataclasses.dataclass
class MemoryRepository(Generic[Entity, EntityData, FallbackKey]):
    memory_data_source: MemoryDataSource
    fallback_data_source: FallbackDataSource[FallbackKey]
    expire_time: int
    entity_cls: ClassVar[Type[Entity]]
    name: ClassVar[str]
    id_name: ClassVar[str]
    key_attrs: ClassVar[Sequence[str]]
    many_key_attrs: ClassVar[Sequence[str]]
    __skip_cls_validation__: Sequence[str] = ()

    def __init_subclass__(
        cls,
        entity_cls: Optional[Type[Entity]] = None,
        name: Optional[str] = None,
        id_name: Optional[str] = None,
        key_attrs: Optional[Sequence[str]] = None,
        many_key_attrs: Optional[Sequence[str]] = None,
    ):
        super().__init_subclass__()

        if cls.__name__ not in cls.__skip_cls_validation__:
            entity_cls = entity_cls or getattr(cls, 'entity_cls', None)
            name = name or getattr(cls, 'name', None)
            id_name = id_name or getattr(cls, 'id_name', 'id')
            key_attrs = key_attrs or getattr(cls, 'key_attrs', None)
            many_key_attrs = many_key_attrs or getattr(
                cls, 'many_key_attrs', None
            )

            if name is None:
                name = re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()
                name = name.replace('_repository', '')

            if key_attrs is None:
                key_attrs = (id_name,)

            has_entity_type = any(
                [
                    cls.get_entity_type != base.get_entity_type  # type: ignore
                    for base in cls.__bases__
                ]
            )

            if not has_entity_type and not entity_cls:
                raise RequiredClassAttributeError(
                    cls.__name__, 'entity_cls or get_entity_type'
                )

            cls.entity_cls = entity_cls
            cls.name = name
            cls.id_name = id_name
            cls.key_attrs = key_attrs
            cls.many_key_attrs = many_key_attrs or cls.key_attrs

    async def get_memory_data(
        self, key: str, query: 'BaseQuery[Entity, EntityData, FallbackKey]',
    ) -> Optional[EntityData]:
        raise NotImplementedError()  # pragma: no cover

    async def get_fallback_data(
        self,
        query: Union['BaseQuery[Entity, EntityData, FallbackKey]', Entity],
        *,
        for_memory: bool = False,
    ) -> Optional[EntityData]:
        raise NotImplementedError()  # pragma: no cover

    def make_entity(
        self,
        data: EntityData,
        query: 'BaseQuery[Entity, EntityData, FallbackKey]',
    ) -> Entity:
        raise NotImplementedError()  # pragma: no cover

    def make_entity_from_fallback(
        self,
        data: EntityData,
        query: 'BaseQuery[Entity, EntityData, FallbackKey]',
    ) -> Entity:
        raise NotImplementedError()  # pragma: no cover

    async def add_memory_data(self, key: str, data: EntityData) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def add_memory_data_from_fallback(
        self,
        key: str,
        query: Union['BaseQuery[Entity, EntityData, FallbackKey]', Entity],
        data: EntityData,
    ) -> EntityData:
        raise NotImplementedError()  # pragma: no cover

    def make_memory_data_from_entity(self, entity: Entity) -> EntityData:
        raise NotImplementedError()  # pragma: no cover

    async def add_fallback(
        self,
        entity: Entity,
        *entities: Entity,
        fallback_ttl: Optional[int] = None,
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def entity(
        self, query: 'Query[Entity, EntityData, FallbackKey]',
    ) -> Entity:
        if query.memory:
            return await self.get_memory(query)
        else:
            return await self.get_fallback(query)

    def entities(
        self, query: 'QueryMany[Entity, EntityData, FallbackKey]',
    ) -> AsyncGenerator[Entity, None]:
        if query.memory:
            return self.get_memory_many(query)
        else:
            return self.get_fallback_many(query)

    async def get_fallback_many(
        self, query: 'QueryMany[Entity, EntityData, FallbackKey]',
    ) -> AsyncGenerator[Entity, None]:
        for query_ in query.queries:
            try:
                yield await self.get_fallback(query_)
            except EntityNotFoundError:
                continue

    async def get_memory_many(
        self, query: 'QueryMany[Entity, EntityData, FallbackKey]',
    ) -> AsyncGenerator[Entity, None]:
        for query_i in query.queries:
            memory_key = self.memory_key(query_i)
            memory_data = await self.get_memory_data(memory_key, query_i)

            if memory_data:
                yield self.make_entity(memory_data, query_i)

            elif not await self.already_got_not_found(query_i):
                fallback_data = await self.get_fallback_data(
                    query_i, for_memory=True
                )

                if fallback_data is None:
                    await self.set_fallback_not_found(query_i)
                else:
                    memory_data = await self.add_memory_data_from_fallback(
                        memory_key, query_i, fallback_data
                    )
                    await self.set_expire_time(memory_key)
                    yield self.make_entity(memory_data, query_i)

    async def get_memory(
        self, query: 'Query[Entity, EntityData, FallbackKey]',
    ) -> Entity:
        memory_key = self.memory_key(query)
        memory_data = await self.get_memory_data(memory_key, query)

        if not memory_data and not await self.already_got_not_found(query):
            fallback_data = await self.get_fallback_data(
                query, for_memory=True
            )

            if fallback_data is None:
                await self.set_fallback_not_found(query)
            else:
                memory_data = await self.add_memory_data_from_fallback(
                    memory_key, query, fallback_data
                )
                await self.set_expire_time(memory_key)

        if not memory_data:
            raise EntityNotFoundError(query)

        return self.make_entity(memory_data, query)

    async def get_fallback(
        self,
        query: Union['BaseQuery[Entity, EntityData, FallbackKey]', Entity],
    ) -> Entity:
        data = await self.get_fallback_data(query)

        if data is None:
            raise EntityNotFoundError(query)

        return self.make_entity_from_fallback(data, query)  # type: ignore

    async def add(
        self,
        *entities: Entity,
        memory: bool = True,
        query: Optional['BaseQuery[Entity, EntityData, FallbackKey]'] = None,
        fallback_ttl: Optional[int] = None,
    ) -> None:
        if memory:
            return await self.add_memory(*entities, fallback_ttl=fallback_ttl)
        else:
            return await self.add_fallback(
                *entities, fallback_ttl=fallback_ttl
            )

    async def add_memory(
        self,
        entity: Entity,
        *entities: Entity,
        fallback_ttl: Optional[int] = None,
    ) -> None:
        memory_key = self.memory_key(entity)

        if await self.memory_data_source.exists(memory_key):
            memory_data = self.make_memory_data_from_entity(entity)
            await self.add_memory_data(memory_key, memory_data)
            await self.set_expire_time(memory_key)

        await self.add_fallback(entity, fallback_ttl=fallback_ttl)
        await self.delete_fallback_not_found(entity)

    async def set_expire_time(self, key: str) -> None:
        await self.memory_data_source.expire(key, self.expire_time)

    def query(
        self,
        *args: Any,
        memory: bool = True,
        many: Optional[Sequence[Any]] = None,
        **kwargs: Any,
    ) -> 'BaseQuery[Entity, EntityData, FallbackKey]':
        return self.make_query(memory=memory, many=many, *args, **kwargs)

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> 'BaseQuery[Entity, EntityData, FallbackKey]':
        return query_factory(self, *args, **kwargs)

    async def delete(
        self, query: 'Query[Entity, EntityData, FallbackKey]',
    ) -> None:
        if query.memory:
            await self.memory_data_source.delete(self.memory_key(query))
            await self.set_fallback_not_found(query)

        await self.fallback_data_source.delete(self.fallback_key(query))

    async def exists(
        self, query: 'Query[Entity, EntityData, FallbackKey]',
    ) -> bool:
        if query.memory:
            memory_key = self.memory_key(query)
            entity_exists = await self.memory_data_source.exists(memory_key)

            if not entity_exists:
                if not await self.already_got_not_found(query):
                    try:
                        await self.get_memory(query)
                    except EntityNotFoundError:
                        return False
                else:
                    return False

        else:
            try:
                await self.get_fallback(query)
            except EntityNotFoundError:
                return False

        return True

    async def already_got_not_found(
        self, query: Union['Query[Entity, EntityData, FallbackKey]', Entity],
    ) -> bool:
        return bool(
            await self.memory_data_source.exists(
                self.fallback_not_found_key(query)
            )
        )

    async def delete_fallback_not_found(
        self, query: Union['Query[Entity, EntityData, FallbackKey]', Entity],
    ) -> None:
        await self.memory_data_source.delete(
            self.fallback_not_found_key(query)
        )

    async def set_fallback_not_found(
        self, query: Union['Query[Entity, EntityData, FallbackKey]', Entity],
    ) -> None:
        key = self.fallback_not_found_key(query)
        await self.memory_data_source.set(key, '1')
        await self.set_expire_time(key)

    def memory_key(
        self, query: 'Union[Query[Entity, EntityData, FallbackKey], Entity]',
    ) -> str:
        if isinstance(query, Query):
            return self.memory_data_source.make_key(
                self.name, *query.key_parts
            )

        elif isinstance(self.get_entity_type(query), _TypedDictMeta):
            return self.memory_data_source.make_key(
                self.name, *self.key_parts(query)
            )

        elif isinstance(query, self.get_entity_type(query)):
            return self.memory_data_source.make_key(
                self.name, *self.key_parts(query)
            )

        raise InvalidQueryError(query)

    def fallback_key(
        self, query: 'Union[Query[Entity, EntityData, FallbackKey], Entity]',
    ) -> FallbackKey:
        if isinstance(query, Query):
            return self.fallback_data_source.make_key(
                self.name, *query.key_parts
            )

        elif isinstance(self.get_entity_type(query), _TypedDictMeta):
            return self.fallback_data_source.make_key(
                self.name, *self.key_parts(query)
            )

        elif isinstance(query, self.get_entity_type(query)):
            return self.fallback_data_source.make_key(
                self.name, *self.key_parts(query)
            )

        raise InvalidQueryError(query)

    @classmethod
    def key_parts(cls, entity: Entity) -> List[Any]:
        if hasattr(entity, '__getitem__'):
            try:
                return [
                    entity[attr_name]  # type: ignore
                    for attr_name in cls.key_attrs
                ]
            except KeyError as error:
                raise InvalidKeyAttributeError(
                    cls.__name__, cls.key_attrs, *error.args,
                )

        try:
            return [getattr(entity, attr_name) for attr_name in cls.key_attrs]
        except AttributeError as error:
            raise InvalidKeyAttributeError(
                cls.__name__,
                cls.key_attrs,
                type(entity).__name__,
                *error.args,
            )

    def fallback_not_found_key(
        self, query: 'Union[Query[Entity, EntityData, FallbackKey], Entity]',
    ) -> str:
        if isinstance(query, Query):
            return self.memory_data_source.make_key(
                self.name, 'not-found', *query.key_parts
            )

        elif isinstance(self.get_entity_type(query), _TypedDictMeta):
            return self.memory_data_source.make_key(
                self.name, 'not-found', *self.key_parts(query)
            )

        elif isinstance(query, self.get_entity_type(query)):
            return self.memory_data_source.make_key(
                self.name, 'not-found', *self.key_parts(query)
            )

        raise InvalidQueryError(query)

    def get_entity_type(
        self,
        query: 'Union[BaseQuery[Entity, EntityData, FallbackKey], Entity]',
    ) -> Type[Entity]:
        return self.entity_cls


def task_done_callback(f: Any) -> None:
    try:
        f.result()
    except EntityNotFoundError:
        ...


from .query import BaseQuery, Query, QueryMany  # noqa isort:skip
from .query import make as query_factory  # noqa isort:skip



from typing import (
    Any,
    ClassVar,
    Dict,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    Union,
)

from .. import DataSource


class GeoPoint(Protocol):
    longitude: float
    latitude: float

    def __init__(self, longitude: float, latitude: float):
        ...


class GeoMember(Protocol):
    member: bytes
    dist: Optional[float]
    coord: Optional[GeoPoint]

    def __init__(
        self,
        member: bytes,
        dist: Optional[float],
        coord: Optional[GeoPoint],
        hash: Optional[str] = None,
    ):
        ...


GeoRadiusOutput = Union[Sequence[GeoMember], Sequence[str]]


RangeOutput = Union[
    Sequence[Tuple[str, float]],
    Sequence[Tuple[bytes, float]],
    Sequence[Union[bytes, str]],
]


class MemoryMultiExec:
    def delete(self, key: str) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    async def execute(self, *, return_exceptions: bool = False) -> Any:
        raise NotImplementedError()  # pragma: no cover


class MemoryDataSource(DataSource):
    key_separator: ClassVar[str] = ':'
    geopoint_cls: ClassVar[Type[GeoPoint]]
    geomember_cls: ClassVar[Type[GeoMember]]

    def make_key(self, *key_parts: str) -> str:
        return self.key_separator.join(key_parts)

    async def get(self, key: str) -> Optional[bytes]:
        raise NotImplementedError()  # pragma: no cover

    async def set(self, key: str, data: str) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def delete(self, key: str) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def expire(self, key: str, time: int) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def exists(self, key: str) -> int:
        raise NotImplementedError()  # pragma: no cover

    async def zrevrange(
        self, key: str, start: int, stop: int, withscores: bool = False
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zrange(
        self,
        key: str,
        start: int = 0,
        stop: int = -1,
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zrevrangebyscore(
        self,
        key: str,
        max: float = float('inf'),
        min: float = float('-inf'),
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zrangebyscore(
        self,
        key: str,
        min: float = float('-inf'),
        max: float = float('inf'),
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        raise NotImplementedError()  # pragma: no cover

    async def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def zcard(self, key: str) -> int:
        raise NotImplementedError()  # pragma: no cover

    async def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def hmget(
        self, key: str, field: Union[str, bytes], *fields: Union[str, bytes]
    ) -> Sequence[Optional[bytes]]:
        raise NotImplementedError()  # pragma: no cover

    async def hgetall(self, key: str) -> Dict[bytes, bytes]:
        raise NotImplementedError()  # pragma: no cover

    def close(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def wait_closed(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    def multi_exec(self) -> MemoryMultiExec:
        raise NotImplementedError()  # pragma: no cover

    async def georadius(
        self,
        key: str,
        longitude: float,
        latitude: float,
        radius: float,
        unit: str = 'm',
        *,
        with_dist: bool = False,
        with_coord: bool = False,
        count: Optional[int] = None,
    ) -> GeoRadiusOutput:
        raise NotImplementedError()  # pragma: no cover

    async def geoadd(
        self,
        key: str,
        longitude: float,
        latitude: float,
        member: Union[str, bytes],
        *args: Any,
        **kwargs: Any,
    ) -> int:
        raise NotImplementedError()  # pragma: no cover



from dbdaora.data_sources.fallback.mongodb import Key

from . import SortedSetRepository


class MongodbSortedSetRepository(SortedSetRepository[Key]):
    __skip_cls_validation__ = ('MongodbSortedSetRepository',)



import dataclasses
from typing import Any, Dict, Optional, Sequence, Union

from . import MemoryDataSource, RangeOutput


@dataclasses.dataclass
class DictMemoryDataSource(MemoryDataSource):
    db: Dict[str, Any] = dataclasses.field(default_factory=dict)

    async def set(self, key: str, data: str) -> None:
        self.db[key] = data.encode()

    async def delete(self, key: str) -> None:
        self.db.pop(key, None)

    async def expire(self, key: str, time: int) -> None:
        ...

    async def exists(self, key: str) -> bool:
        return key in self.db

    async def zrange(
        self,
        key: str,
        start: int = 0,
        stop: int = -1,
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        data: Optional[RangeOutput] = self.db.get(key)

        if data is None:
            return None

        return [i[0] for i in self.db[key]]

    async def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> None:
        data = [score, member] + list(pairs)
        self.db[key] = sorted(
            [
                (
                    data[i].encode() if isinstance(data[i], str) else data[i],  # type: ignore
                    data[i - 1],
                )
                for i in range(1, len(data), 2)
            ],
            key=lambda d: d[1],
        )

    async def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> None:
        data = [field, value] + list(pairs)
        self.db[key] = {
            f.encode()
            if isinstance(f := data[i - 1], str)  # noqa
            else (f if isinstance(f, bytes) else str(f).encode()): v.encode()
            if isinstance(v := data[i], str)  # noqa
            else (v if isinstance(v, bytes) else str(v).encode())
            for i in range(1, len(data), 2)
        }

    async def hmget(
        self, key: str, field: Union[str, bytes], *fields: Union[str, bytes]
    ) -> Sequence[Optional[bytes]]:
        data: Dict[bytes, Any] = self.db.get(key, {})

        return [
            None
            if (d := data.get(f.encode() if isinstance(f, str) else f))  # noqa
            is None
            else (
                d
                if isinstance(d, bytes)
                else (d.encode() if isinstance(d, str) else str(d).encode())
            )
            for f in (field,) + fields
        ]

    async def hgetall(self, key: str) -> Dict[bytes, bytes]:
        return {
            f: d.encode()
            if isinstance(d, str)
            else (
                d
                if isinstance(d, bytes)
                else (d.encode() if isinstance(d, str) else str(d).encode())
            )
            for f, d in self.db.get(key, {}).items()
        }



"""Communicates with databases using repository pattern and service patterns"""


__version__ = '0.21.0'

from dbdaora.cache import CacheType, TTLDaoraCache
from dbdaora.circuitbreaker import AsyncCircuitBreaker
from dbdaora.data_sources.fallback import FallbackDataSource
from dbdaora.data_sources.fallback.dict import DictFallbackDataSource
from dbdaora.data_sources.memory import MemoryDataSource
from dbdaora.data_sources.memory.dict import DictMemoryDataSource
from dbdaora.exceptions import EntityNotFoundError, InvalidGeoSpatialDataError
from dbdaora.geospatial.entity import GeoSpatialData, GeoSpatialEntity
from dbdaora.geospatial.factory import make_service as make_geospatial_service
from dbdaora.geospatial.query import GeoSpatialQuery
from dbdaora.geospatial.repositories import GeoSpatialRepository
from dbdaora.geospatial.service import GeoSpatialService
from dbdaora.hash.factory import make_service as make_hash_service
from dbdaora.hash.query import HashQuery, HashQueryMany
from dbdaora.hash.repositories import HashData, HashRepository
from dbdaora.hash.service import HashService
from dbdaora.hashring import HashRing
from dbdaora.query import Query, QueryMany
from dbdaora.repository import MemoryRepository
from dbdaora.service import Service
from dbdaora.service.builder import build as build_service
from dbdaora.service.builder import build_cache
from dbdaora.sorted_set.entity import (
    SortedSetData,
    SortedSetDictEntity,
    SortedSetEntity,
)
from dbdaora.sorted_set.factory import make_service as make_sorted_set_service
from dbdaora.sorted_set.query import SortedSetQuery
from dbdaora.sorted_set.repositories import SortedSetRepository


from dbdaora.boolean.repositories import BooleanRepository  # noqa isort:skip
from dbdaora.boolean.service import BooleanService  # noqa isort:skip
from dbdaora.boolean.repositories.datastore import (  # noqa isort:skip
    DatastoreBooleanRepository,
)

try:
    from dbdaora.data_sources.fallback.datastore import (
        DatastoreDataSource,
        KindKeyDatastoreDataSource,
    )
    from dbdaora.hash.repositories.datastore import DatastoreHashRepository
    from dbdaora.sorted_set.repositories.datastore import (
        DatastoreSortedSetRepository,
    )
    from dbdaora.hash.service.datastore import DatastoreHashService
    from dbdaora.geospatial.service.datastore import DatastoreGeoSpatialService
    from dbdaora.geospatial.repositories.datastore import (
        DatastoreGeoSpatialRepository,
    )
    from dbdaora.sorted_set.service.datastore import DatastoreSortedSetService
except ImportError:
    DatastoreDataSource = None  # type: ignore
    KindKeyDatastoreDataSource = None  # type: ignore
    DatastoreHashRepository = None  # type: ignore
    DatastoreSortedSetRepository = None  # type: ignore
    DatastoreHashService = None  # type: ignore
    DatastoreGeoSpatialService = None  # type: ignore
    DatastoreGeoSpatialRepository = None  # type: ignore
    DatastoreSortedSetService = None  # type: ignore

try:
    from dbdaora.data_sources.memory.aioredis import (
        AioRedisDataSource,
        ShardsAioRedisDataSource,
        make as make_aioredis_data_source,
    )
except ImportError:
    AioRedisDataSource = None  # type: ignore
    ShardsAioRedisDataSource = None  # type: ignore
    make_aioredis_data_source = None  # type: ignore

try:
    from dbdaora.data_sources.fallback.mongodb import (
        MongoDataSource,
        Key as MongoKey,
        CollectionKeyMongoDataSource,
    )
    from dbdaora.hash.repositories.mongodb import MongodbHashRepository
    from dbdaora.hash.service.mongodb import MongoHashService
    from dbdaora.geospatial.service.mongodb import MongoGeoSpatialService
    from dbdaora.geospatial.repositories.mongodb import (
        MongodbGeoSpatialRepository,
    )
    from dbdaora.sorted_set.service.mongodb import MongoSortedSetService
    from dbdaora.sorted_set.repositories.mongodb import (
        MongodbSortedSetRepository,
    )
except ImportError:
    MongoDataSource = None  # type: ignore
    CollectionKeyMongoDataSource = None  # type: ignore
    MongodbHashRepository = None  # type: ignore
    MongoGeoSpatialService = None  # type: ignore
    MongodbGeoSpatialRepository = None  # type: ignore
    MongoHashService = None  # type: ignore
    MongoSortedSetService = None  # type: ignore
    MongodbSortedSetRepository = None  # type: ignore


__all__ = [
    'MemoryRepository',
    'HashRepository',
    'HashQuery',
    'HashData',
    'SortedSetRepository',
    'SortedSetQuery',
    'SortedSetEntity',
    'DictFallbackDataSource',
    'HashService',
    'AsyncCircuitBreaker',
    'HashRing',
    'FallbackDataSource',
    'MemoryDataSource',
    'DictMemoryDataSource',
    'build_service',
    'CacheType',
    'Service',
    'EntityNotFoundError',
    'HashQueryMany',
    'make_hash_service',
    'SortedSetData',
    'SortedSetDictEntity',
    'TTLDaoraCache',
    'build_cache',
    'BooleanRepository',
    'DatastoreBooleanRepository',
    'BooleanService',
    'Query',
    'QueryMany',
    'InvalidGeoSpatialDataError',
    'GeoSpatialQuery',
    'GeoSpatialEntity',
    'GeoSpatialService',
    'GeoSpatialRepository',
    'make_geospatial_service',
    'GeoSpatialData',
    'make_sorted_set_service',
]

if AioRedisDataSource:
    __all__.append('AioRedisDataSource')

if ShardsAioRedisDataSource:
    __all__.append('ShardsAioRedisDataSource')

if make_aioredis_data_source:
    __all__.append('make_aioredis_data_source')

if DatastoreDataSource:
    __all__.append('DatastoreDataSource')

if KindKeyDatastoreDataSource:
    __all__.append('KindKeyDatastoreDataSource')

if DatastoreHashRepository:
    __all__.append('DatastoreHashRepository')

if DatastoreSortedSetRepository:
    __all__.append('DatastoreSortedSetRepository')

if MongoDataSource:
    __all__.append('MongoDataSource')

if CollectionKeyMongoDataSource:
    __all__.append('CollectionKeyMongoDataSource')

if MongodbHashRepository:
    __all__.append('MongodbHashRepository')

if MongoHashService:
    __all__.append('MongoHashService')

if DatastoreHashService:
    __all__.append('DatastoreHashService')

if MongoKey:
    __all__.append('MongoKey')

if MongodbGeoSpatialRepository:
    __all__.append('MongodbGeoSpatialRepository')

if MongoGeoSpatialService:
    __all__.append('MongoGeoSpatialService')

if DatastoreGeoSpatialRepository:
    __all__.append('DatastoreGeoSpatialRepository')

if DatastoreGeoSpatialService:
    __all__.append('DatastoreGeoSpatialService')

if DatastoreSortedSetService:
    __all__.append('DatastoreSortedSetService')

if MongodbSortedSetRepository:
    __all__.append('MongodbSortedSetRepository')

if MongoSortedSetService:
    __all__.append('MongoSortedSetService')



from typing import Any

from google.cloud.datastore import Entity, Key
from jsondaora.serializers import OrjsonDefaultTypes

from . import BooleanRepository


OrjsonDefaultTypes.types_default_map[Entity] = lambda e: dict(**e)


class DatastoreBooleanRepository(BooleanRepository[Key]):
    __skip_cls_validation__ = ('DatastoreBooleanRepository',)

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await super().add_fallback(
            entity, *entities, exclude_from_indexes=('value',)
        )



from dbdaora.data_sources.fallback.mongodb import Key

from . import GeoSpatialRepository


class MongodbGeoSpatialRepository(GeoSpatialRepository[Key]):
    __skip_cls_validation__ = ('MongodbGeoSpatialRepository',)



from typing import Any, ClassVar, Optional, Sequence, Type, get_type_hints

from google.cloud.datastore import Entity, Key
from jsondaora.serializers import OrjsonDefaultTypes

from . import HashRepository


OrjsonDefaultTypes.types_default_map[Entity] = lambda e: dict(**e)


class DatastoreHashRepository(HashRepository[Key]):
    __skip_cls_validation__ = ('DatastoreHashRepository',)
    exclude_from_indexes: ClassVar[Sequence[str]] = ()
    exclude_all_from_indexes: ClassVar[bool] = False

    def __init_subclass__(
        cls,
        entity_cls: Optional[Type[Any]] = None,
        name: Optional[str] = None,
        id_name: Optional[str] = None,
        key_attrs: Optional[Sequence[str]] = None,
        many_key_attrs: Optional[Sequence[str]] = None,
    ):
        super().__init_subclass__(
            entity_cls, name, id_name, key_attrs, many_key_attrs,
        )

        if cls.exclude_all_from_indexes:
            cls.exclude_from_indexes = tuple(
                get_type_hints(cls.entity_cls).keys()
            )

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await super().add_fallback(
            entity, *entities, exclude_from_indexes=self.exclude_from_indexes
        )



from typing import Any

from google.cloud.datastore import Key

from . import SortedSetRepository


class DatastoreSortedSetRepository(SortedSetRepository[Key]):
    __skip_cls_validation__ = ('DatastoreSortedSetRepository',)

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await super().add_fallback(
            entity, *entities, exclude_from_indexes=('values',)
        )



import dataclasses
from functools import partial
from typing import List

import pytest
from aioredis import GeoMember, GeoPoint, RedisError

from dbdaora import (
    DictFallbackDataSource,
    GeoSpatialRepository,
    GeoSpatialService,
    build_service,
    make_aioredis_data_source,
)


@pytest.mark.asyncio
@pytest.fixture
async def fake_service(mocker, fallback_data_source, fake_repository_cls):
    memory_data_source_factory = partial(
        make_aioredis_data_source,
        'redis://',
        'redis://localhost/1',
        'redis://localhost/2',
    )

    async def fallback_data_source_factory():
        return fallback_data_source

    service = await build_service(
        GeoSpatialService,
        fake_repository_cls,
        memory_data_source_factory,
        fallback_data_source_factory,
        repository_expire_time=1,
        cb_failure_threshold=0,
        cb_recovery_timeout=10,
        cb_expected_exception=RedisError,
        logger=mocker.MagicMock(),
    )

    yield service

    service.repository.memory_data_source.close()
    await service.repository.memory_data_source.wait_closed()


@pytest.fixture
def fallback_data_source():
    return DictFallbackDataSource()


@dataclasses.dataclass
class FakeEntity:
    fake_id: str
    fake2_id: str
    data: List[GeoMember]


class FakeGeoSpatialRepository(GeoSpatialRepository[str]):
    name = 'fake'
    key_attrs = ('fake2_id', 'fake_id')
    entity_cls = FakeEntity


@pytest.fixture
def fake_entity_cls():
    return FakeEntity


@pytest.fixture
def fake_repository_cls():
    return FakeGeoSpatialRepository


@pytest.fixture
def fake_entity(fake_entity_cls):
    return fake_entity_cls(
        fake_id='fake',
        fake2_id='fake2',
        data=[
            GeoMember(
                member=b'm1',
                dist=0.0003,
                hash=None,
                coord=GeoPoint(6.000002324581146, 4.999999830436074),
            ),
            GeoMember(
                member=b'm2',
                dist=0.0003,
                hash=None,
                coord=GeoPoint(6.000002324581146, 4.999999830436074),
            ),
        ],
    )


@pytest.fixture
def fake_entity_add(fake_entity_cls):
    return fake_entity_cls(
        fake_id='fake',
        fake2_id='fake2',
        data=GeoMember(
            member=b'm1',
            dist=0.0003,
            hash=None,
            coord=GeoPoint(6.000002324581146, 4.999999830436074),
        ),
    )


@pytest.fixture
def fake_entity_add2(fake_entity_cls):
    return fake_entity_cls(
        fake_id='fake',
        fake2_id='fake2',
        data=GeoMember(
            member=b'm2',
            dist=0.0003,
            hash=None,
            coord=GeoPoint(6.000002324581146, 4.999999830436074),
        ),
    )


@pytest.fixture
def serialized_fake_entity():
    return [
        (6.000002324581146, 4.999999830436074, 'm1'),
        (6.000002324581146, 4.999999830436074, 'm2'),
    ]


@pytest.fixture
def fake_fallback_data_entity(fake_entity):
    return {
        'latitude': fake_entity.data[0].coord.latitude,
        'longitude': fake_entity.data[0].coord.longitude,
        'member': fake_entity.data[0].member,
    }


@pytest.fixture
def fake_fallback_data_entity2(fake_entity):
    return {
        'latitude': fake_entity.data[1].coord.latitude,
        'longitude': fake_entity.data[1].coord.longitude,
        'member': fake_entity.data[1].member,
    }


@pytest.mark.asyncio
@pytest.fixture
async def memory_data_source():
    return await make_aioredis_data_source(
        'redis://', 'redis://localhost/1', 'redis://localhost/2'
    )


@pytest.mark.asyncio
@pytest.fixture
async def repository(
    mocker, memory_data_source, fallback_data_source, fake_repository_cls
):
    yield fake_repository_cls(
        memory_data_source=memory_data_source,
        fallback_data_source=fallback_data_source,
        expire_time=1,
    )
    memory_data_source.close()
    await memory_data_source.wait_closed()



import dataclasses
import random
import time
from enum import Enum
from typing import Any, Dict, Optional, Tuple, Type, Union

from cachetools import Cache, LFUCache, LRUCache, TTLCache


@dataclasses.dataclass(init=False)
class TTLDaoraCache:
    maxsize: int
    ttl: int
    ttl_failure_threshold: int
    cache: Dict[str, Tuple[Any, float]]
    first_set_time: float
    clean_keys_size: int

    def __init__(
        self,
        maxsize: int,
        ttl: int,
        ttl_failure_threshold: int = 0,
        cache: Optional[Dict[str, Tuple[Any, float]]] = None,
        first_set_time: float = 0,
    ):
        if cache is None:
            cache = {}

        self.maxsize = maxsize
        self.ttl = ttl
        self.ttl_failure_threshold = ttl_failure_threshold
        self.cache = cache
        self.first_set_time = first_set_time
        self.clean_keys_size = int(self.maxsize * 0.1)

    def __setitem__(self, key: str, data: Any) -> None:
        if len(self.cache) < self.maxsize:
            set_time = time.time() - self.ttl_threshold
            self.cache[key] = (data, set_time)

            if self.first_set_time <= 0:
                self.first_set_time = set_time

        elif self.first_set_time + self.ttl < time.time():
            self.first_set_time = 0

            for i, key_to_pop in enumerate(self.cache.keys()):
                self.cache.pop(key_to_pop, None)

                if i >= self.clean_keys_size:
                    break

            set_time = time.time() - self.ttl_threshold
            self.cache[key] = (data, set_time)

    def get(self, key: str, default: Any = None) -> Any:
        data = self.cache.get(key)

        if data is not None:
            if data[1] + self.ttl >= time.time():
                return data[0]

            else:
                self.cache.pop(key, None)
                return data[0]

        return default

    @property
    def ttl_threshold(self) -> int:
        if self.ttl_failure_threshold:
            return random.randint(0, self.ttl_failure_threshold)

        return 0


class CacheType(Enum):
    value: Union[Type[Cache]]

    TTL = TTLCache
    LFU = LFUCache
    LRU = LRUCache
    TTLDAORA = TTLDaoraCache



from logging import Logger, getLogger
from typing import Any, AsyncGenerator, Optional, Sequence, Union

from cachetools import Cache

from ...circuitbreaker import AsyncCircuitBreaker
from ...entity import Entity
from ...exceptions import CacheNotAvailableError
from ...keys import FallbackKey
from ...repository import MemoryRepository
from ...service import CacheAlreadyNotFound, Service
from ..entity import GeoSpatialData


class GeoSpatialService(Service[Entity, GeoSpatialData, FallbackKey]):
    def __init__(
        self,
        repository: MemoryRepository[Entity, GeoSpatialData, FallbackKey],
        circuit_breaker: AsyncCircuitBreaker,
        fallback_circuit_breaker: AsyncCircuitBreaker,
        cache: Optional[Cache] = None,
        exists_cache: Optional[Cache] = None,
        logger: Logger = getLogger(__name__),
    ):
        if cache is not None:
            raise CacheNotAvailableError(
                'GeoSpatialService class do not support cache'
            )

        super().__init__(
            repository=repository,
            circuit_breaker=circuit_breaker,
            fallback_circuit_breaker=fallback_circuit_breaker,
            cache=None,
            exists_cache=None,
            logger=logger,
        )
        self.entity_circuit = self.repository.entity
        self.entities_circuit = self.repository.entities

    def get_many(
        self, *ids: str, **filters: Any,
    ) -> AsyncGenerator[Entity, None]:
        raise NotImplementedError()  # pragma: no cover

    def get_many_cached(
        self,
        ids: Sequence[str],
        cache: Cache,
        memory: bool = True,
        **filters: Any,
    ) -> AsyncGenerator[Entity, None]:
        raise NotImplementedError()  # pragma: no cover

    def get_cached_entity(
        self, id: str, key_suffix: str, **filters: Any,
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    def cache_key(self, id: str, suffix: str) -> str:
        raise NotImplementedError()  # pragma: no cover

    def set_cached_entity(
        self,
        id: str,
        key_suffix: str,
        entity: Union[Entity, CacheAlreadyNotFound],
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    def cache_key_suffix(self, **filters: Any) -> str:
        raise NotImplementedError()  # pragma: no cover

    async def get_one_cached(
        self, cache: Cache, memory: bool = True, **filters: Any,
    ) -> Any:
        raise NotImplementedError()  # pragma: no cover

    async def delete(
        self, entity_id: Optional[str] = None, **filters: Any
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def exists(self, id: Optional[str] = None, **filters: Any) -> bool:
        raise NotImplementedError()  # pragma: no cover

    async def exists_cached(
        self, cache: Cache, memory: bool = True, **filters: Any,
    ) -> bool:
        raise NotImplementedError()  # pragma: no cover



from datetime import datetime
from typing import (
    Any,
    Awaitable,
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from circuitbreaker import (
    STATE_CLOSED,
    STATE_OPEN,
    CircuitBreaker,
    CircuitBreakerError,
)


FuncReturn = TypeVar('FuncReturn')


class AsyncCircuitBreaker(CircuitBreaker):
    _last_failure: Optional[Exception]

    def __init__(
        self,
        failure_threshold: Optional[int] = None,
        recovery_timeout: Optional[int] = None,
        expected_exception: Optional[
            Union[Type[Exception], Tuple[Type[Exception], ...]]
        ] = None,
        name: Optional[str] = None,
        fallback_function: Optional[
            Callable[..., Awaitable[FuncReturn]]
        ] = None,
    ):
        self._last_failure = None
        self._failure_count = 0
        self._failure_threshold = (
            failure_threshold
            if failure_threshold is not None
            else self.FAILURE_THRESHOLD
        )
        self._recovery_timeout = recovery_timeout or self.RECOVERY_TIMEOUT
        self._expected_exception = (
            expected_exception or self.EXPECTED_EXCEPTION
        )
        self._fallback_function = fallback_function or self.FALLBACK_FUNCTION
        self._name = name
        self._state = STATE_CLOSED
        self._opened = datetime.utcnow()

    async def call(
        self,
        func: Callable[..., Awaitable[FuncReturn]],
        *args: Any,
        **kwargs: Any,
    ) -> FuncReturn:
        if self.opened:
            if self.fallback_function:
                return await self.fallback_function(*args, **kwargs)
            else:
                raise DBDaoraCircuitBreakerError(self, func.__name__)

        try:
            result = await func(*args, **kwargs)
        except self._expected_exception as e:
            self.set_failure(func.__name__, e)
            raise

        self.__call_succeeded()
        return result

    def __call_succeeded(self) -> None:
        self._state = STATE_CLOSED
        self._last_failure = None
        self._failure_count = 0

    def __call_failed(self) -> None:
        self._failure_count += 1
        if self._failure_count >= self._failure_threshold:
            self._state = STATE_OPEN
            self._opened = datetime.utcnow()

    @property
    def expected_exception(
        self,
    ) -> Union[Type[Exception], Tuple[Type[Exception], ...]]:
        return self._expected_exception

    def set_failure(self, func_name: str, error: Exception) -> None:
        self._last_failure = error
        self.__call_failed()

        if self._failure_threshold == 0:
            raise DBDaoraCircuitBreakerError(self, func_name)

    def set_success(self) -> None:
        self.__call_succeeded()


class DBDaoraCircuitBreakerError(CircuitBreakerError):
    _circuit_breaker: AsyncCircuitBreaker
    last_failure: Optional[Exception]

    def __init__(
        self,
        circuit_breaker: CircuitBreaker,
        name_sufix: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(circuit_breaker, *args, **kwargs)
        self._name_sufix = name_sufix
        self.last_failure = self._circuit_breaker._last_failure

    def __str__(self, *args: Any, **kwargs: Any) -> str:
        return (
            'Circuit "%s" OPEN until %s (%d failures, %d sec remaining) (last_failure: %r)'
            % (
                self._circuit_breaker.name
                if self._name_sufix is None
                else f'{self._circuit_breaker.name}_{self._name_sufix}',
                self._circuit_breaker.open_until,
                self._circuit_breaker.failure_count,
                round(self._circuit_breaker.open_remaining),
                self._circuit_breaker.last_failure,
            )
        )



import dataclasses
from typing import Union

from dbdaora.data_sources.memory import GeoMember, GeoRadiusOutput


GeoSpatialData = Union[GeoMember, GeoRadiusOutput]


@dataclasses.dataclass
class GeoSpatialEntity:
    id: str
    data: GeoSpatialData



import dataclasses
from typing import Any, ClassVar, List, Optional, Sequence, Tuple, Type, Union

from dbdaora.keys import FallbackKey
from dbdaora.query import BaseQuery, Query, QueryMany


@dataclasses.dataclass(init=False)
class HashQuery(Query[Any, 'HashData', FallbackKey]):
    repository: 'HashRepository[FallbackKey]'
    fields: Optional[Sequence[str]] = None

    def __init__(
        self,
        repository: 'HashRepository[FallbackKey]',
        *args: Any,
        memory: bool = True,
        key_parts: Optional[List[Any]] = None,
        fields: Optional[Sequence[str]] = None,
        **kwargs: Any,
    ):
        super().__init__(
            repository, memory=memory, key_parts=key_parts, *args, **kwargs,
        )
        self.fields = fields


@dataclasses.dataclass(init=False)
class HashQueryMany(QueryMany[Any, 'HashData', FallbackKey]):
    query_cls: ClassVar[Type[HashQuery[FallbackKey]]] = HashQuery[FallbackKey]
    queries: Sequence[HashQuery[FallbackKey]]  # type: ignore
    repository: 'HashRepository[FallbackKey]'
    fields: Optional[Sequence[str]] = None

    def __init__(
        self,
        repository: 'HashRepository[FallbackKey]',
        *args: Any,
        many: List[Union[Any, Tuple[Any, ...]]],
        memory: bool = True,
        many_key_parts: Optional[List[List[Any]]] = None,
        fields: Optional[Sequence[str]] = None,
        **kwargs: Any,
    ):
        super().__init__(
            repository,
            memory=memory,
            many=many,
            many_key_parts=many_key_parts,
            *args,
            **kwargs,
        )
        self.fields = fields

        for query in self.queries:
            query.fields = fields


def make(*args: Any, **kwargs: Any) -> BaseQuery[Any, 'HashData', FallbackKey]:
    if kwargs.get('many') or kwargs.get('many_key_parts'):
        return HashQueryMany(*args, **kwargs)

    return HashQuery(*args, **kwargs)


from .repositories import HashRepository, HashData  # noqa isort:skip



import dataclasses
import itertools
from typing import Any, Dict, Optional, Sequence, Union

from jsondaora import dataclasses as jdataclasses

from dbdaora.exceptions import InvalidEntityTypeError
from dbdaora.keys import FallbackKey
from dbdaora.query import BaseQuery
from dbdaora.repository import MemoryRepository


HashData = Union[
    Dict[str, Any], Dict[bytes, Any],
]


class HashRepository(MemoryRepository[Any, HashData, FallbackKey]):
    __skip_cls_validation__ = ('HashRepository',)

    async def get_memory_data(  # type: ignore
        self, key: str, query: 'HashQuery[FallbackKey]',
    ) -> Optional[Union[HashData, Dict[bytes, Any]]]:
        data: Optional[Union[HashData, Dict[bytes, Any]]]

        if query.fields:
            data = self.make_hmget_dict(
                query.fields,
                await self.memory_data_source.hmget(key, *query.fields),
            )

            if not data:
                return None

            return data

        data = await self.memory_data_source.hgetall(key)

        if not data:
            return None

        return data

    def make_hmget_dict(
        self, fields: Sequence[str], data: Sequence[Optional[bytes]]
    ) -> Dict[bytes, Any]:
        return {f.encode(): v for f, v in zip(fields, data) if v is not None}

    async def get_fallback_data(  # type: ignore
        self, query: 'HashQuery[FallbackKey]', *, for_memory: bool = False,
    ) -> Optional[Union[HashData]]:
        data = await self.fallback_data_source.get(self.fallback_key(query))

        if data is None:
            return None

        if not for_memory and query.fields:
            return self.make_fallback_data_fields(query, data)

        elif for_memory:
            return {
                k: int(v) if isinstance(v, bool) else v
                for k, v in data.items()
            }

        return data

    def make_fallback_data_fields(
        self, query: 'HashQuery[FallbackKey]', data: HashData,
    ) -> HashData:
        return {f: v for f, v in data.items() if f in query.fields}  # type: ignore

    def make_fallback_data_fields_with_bytes_keys(
        self, query: 'HashQuery[FallbackKey]', data: HashData,
    ) -> HashData:
        return {
            f: v for f, v in data.items() if f.decode() in query.fields  # type: ignore
        }

    def make_entity(
        self, data: HashData, query: 'BaseQuery[Any, HashData, FallbackKey]',
    ) -> Any:
        if dataclasses.is_dataclass(self.get_entity_type(query)):
            return jdataclasses.asdataclass(
                data, self.get_entity_type(query), has_bytes_keys=True
            )

        raise InvalidEntityTypeError(self.get_entity_type(query))

    def make_entity_from_fallback(
        self, data: HashData, query: 'BaseQuery[Any, HashData, FallbackKey]',
    ) -> Any:
        return jdataclasses.asdataclass(data, self.get_entity_type(query))

    async def add_memory_data(
        self, key: str, data: HashData, from_fallback: bool = False
    ) -> None:
        delete_future = self.memory_data_source.delete(key)
        hmset_future = self.memory_data_source.hmset(
            key, *itertools.chain(*data.items())
        )
        await delete_future
        await hmset_future

    async def add_memory_data_from_fallback(
        self,
        key: str,
        query: Union[BaseQuery[Any, HashData, FallbackKey], Any],
        data: HashData,
    ) -> HashData:
        data = self.make_memory_data_from_fallback(query, data)  # type: ignore
        await self.memory_data_source.hmset(
            key, *itertools.chain(*data.items())  # type: ignore
        )

        if isinstance(query, HashQuery) and query.fields:
            return self.make_fallback_data_fields_with_bytes_keys(query, data)

        return data

    def make_memory_data_from_fallback(
        self,
        query: Union[BaseQuery[Any, HashData, FallbackKey], Any],
        data: Dict[str, Any],
    ) -> Dict[bytes, Any]:
        return {
            k.encode(): int(v) if isinstance(v, bool) else v
            for k, v in jdataclasses.asdict(data, dumps_value=True).items()
            if v is not None
        }

    def make_memory_data_from_entity(self, entity: Any) -> HashData:
        return {
            k: int(v) if isinstance(v, bool) else v
            for k, v in jdataclasses.asdict(entity, dumps_value=True).items()
            if v is not None
        }

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        data = jdataclasses.asdict(entity)
        await self.fallback_data_source.put(
            self.fallback_key(entity), data, **kwargs
        )

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> 'BaseQuery[Any, HashData, FallbackKey]':
        return query_factory(self, *args, **kwargs)


from ..query import HashQuery, HashQueryMany  # noqa isort:skip
from ..query import make as query_factory  # noqa isort:skip



import dataclasses
from typing import Any, ClassVar, Dict, Iterable, Optional

from dbdaora.data_sources.fallback import FallbackDataSource


@dataclasses.dataclass
class DictFallbackDataSource(FallbackDataSource[str]):
    db: Dict[Optional[str], Dict[str, Any]] = dataclasses.field(
        default_factory=dict
    )
    key_separator: ClassVar[str] = ':'

    def make_key(self, *key_parts: str) -> str:
        return self.key_separator.join([p for p in key_parts if p])

    async def get(self, key: str) -> Optional[Dict[str, Any]]:
        return self.db.get(key)

    async def put(self, key: str, data: Dict[str, Any], **kwargs: Any) -> None:
        self.db[key] = data

    async def delete(self, key: str) -> None:
        self.db.pop(key, None)

    async def query(self, key: str, **kwargs: Any) -> Iterable[Dict[str, Any]]:
        return self.db.values()



import dataclasses
import datetime
from typing import Any, ClassVar, Dict, Iterable, Optional, Set

import motor.motor_asyncio as motor
from pymongo.errors import OperationFailure

from . import FallbackDataSource


@dataclasses.dataclass
class Key:
    collection_name: str
    document_id: str


@dataclasses.dataclass
class MongoDataSource(FallbackDataSource[Key]):
    database_name: str
    client: motor.AsyncIOMotorClient = dataclasses.field(
        default_factory=motor.AsyncIOMotorClient
    )
    collections_has_ttl_index: ClassVar[Set[str]] = set()

    def make_key(self, *key_parts: Any) -> Key:
        return Key(
            key_parts[0],
            self.key_separator.join([str(k) for k in key_parts[1:]]),
        )

    async def get(self, key: Key) -> Optional[Dict[str, Any]]:
        collection = self.collection(key)
        return await collection.find_one({'_id': key.document_id})

    async def put(
        self,
        key: Key,
        data: Dict[str, Any],
        exclude_from_indexes: Iterable[str] = (),
        **kwargs: Any,
    ) -> None:
        document_ttl = kwargs.get('fallback_ttl')

        if document_ttl:
            data['last_modified'] = datetime.datetime.now()

            if key.collection_name not in type(self).collections_has_ttl_index:
                try:
                    await self.create_ttl_index(key, document_ttl)
                except OperationFailure:
                    if await self.drop_ttl_index(key, document_ttl):
                        await self.create_ttl_index(key, document_ttl)

                type(self).collections_has_ttl_index.add(key.collection_name)

        collection = self.collection(key)
        await collection.replace_one(
            {'_id': key.document_id}, data, upsert=True,
        )

    async def delete(self, key: Key) -> None:
        collection = self.collection(key)
        await collection.delete_one({'_id': key.document_id})

    def collection(self, key: Key) -> motor.AsyncIOMotorCollection:
        return self.client[self.database_name][key.collection_name]

    async def query(self, key: Key, **kwargs: Any) -> Iterable[Dict[str, Any]]:
        return [i async for i in self.collection(key).find(**kwargs)]

    async def create_ttl_index(self, key: Key, document_ttl: int) -> None:
        await self.collection(key).create_index(
            'last_modified', expireAfterSeconds=document_ttl,
        )

    async def drop_ttl_index(self, key: Key, document_ttl: int) -> bool:
        collection = self.collection(key)

        async for index in collection.list_indexes():
            if 'last_modified' in index['key']:
                await collection.drop_index(index['name'])
                return True

        return False


class CollectionKeyMongoDataSource(MongoDataSource):
    def make_key(self, *key_parts: Any) -> Key:
        return Key(
            self.key_separator.join([str(k) for k in key_parts[:-1]]),
            key_parts[-1],
        )



import dataclasses
from typing import Any, List, Optional

from dbdaora.keys import FallbackKey
from dbdaora.query import Query

from .entity import Entity, SortedSetData


@dataclasses.dataclass
class SortedSetQuery(Query[Entity, SortedSetData, FallbackKey]):
    repository: 'SortedSetRepository[FallbackKey]'
    reverse: bool = False
    withscores: bool = False
    page: Optional[int] = None
    page_size: Optional[int] = None
    min_score: Optional[float] = None
    max_score: Optional[float] = None
    withmaxsize: bool = False

    def __init__(
        self,
        repository: 'SortedSetRepository[FallbackKey]',
        *args: Any,
        memory: bool = True,
        key_parts: Optional[List[Any]] = None,
        reverse: bool = False,
        withscores: bool = False,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        min_score: Optional[float] = None,
        max_score: Optional[float] = None,
        withmaxsize: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            repository, memory=memory, key_parts=key_parts, *args, **kwargs,
        )
        self.reverse = reverse
        self.withscores = withscores
        self.page = page
        self.page_size = page_size
        self.min_score = min_score
        self.max_score = max_score
        self.withmaxsize = withmaxsize


from .repositories import SortedSetRepository  # noqa isort:skip



from typing import Any, ClassVar, Sequence

from google.cloud.datastore import Entity, Key
from jsondaora.serializers import OrjsonDefaultTypes

from . import GeoSpatialRepository


OrjsonDefaultTypes.types_default_map[Entity] = lambda e: dict(**e)


class DatastoreGeoSpatialRepository(GeoSpatialRepository[Key]):
    __skip_cls_validation__ = ('DatastoreGeoSpatialRepository',)
    exclude_from_indexes: ClassVar[Sequence[str]] = (
        'latitude',
        'longitude',
        'member',
    )

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await super().add_fallback(
            entity, *entities, exclude_from_indexes=self.exclude_from_indexes
        )



import dataclasses
from typing import Any, Dict, Iterable, Optional

from google.cloud.datastore import Client, Entity, Key

from . import FallbackDataSource


@dataclasses.dataclass
class DatastoreDataSource(FallbackDataSource[Key]):
    client: Client = dataclasses.field(default_factory=Client)

    def make_key(self, *key_parts: Any) -> Key:
        return self.client.key(
            key_parts[0],
            self.key_separator.join([str(k) for k in key_parts[1:]]),
        )

    async def get(self, key: Key) -> Optional[Dict[str, Any]]:
        entity = self.client.get(key)
        return None if entity is None else entity_asdict(entity)

    async def put(
        self,
        key: Key,
        data: Dict[str, Any],
        exclude_from_indexes: Iterable[str] = (),
        **kwargs: Any,
    ) -> None:
        entity = Entity(key, exclude_from_indexes=exclude_from_indexes)
        entity.update(data)
        self.client.put(entity)

    async def delete(self, key: Key) -> None:
        self.client.delete(key)

    async def query(self, key: Key, **kwargs: Any) -> Iterable[Dict[str, Any]]:
        return self.client.query(kind=key.kind, **kwargs).fetch()


def entity_asdict(entity: Entity) -> Dict[str, Any]:
    return {
        k: entity_asdict(v) if isinstance(v, Entity) else v
        for k, v in entity.items()
    }


class KindKeyDatastoreDataSource(DatastoreDataSource):
    def make_key(self, *key_parts: Any) -> Key:
        return self.client.key(
            self.key_separator.join([str(k) for k in key_parts[:-1]]),
            key_parts[-1],
        )



import asyncio
import dataclasses
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
)

from aioredis import GeoMember, GeoPoint, Redis, create_redis_pool
from aioredis.commands.transaction import MultiExec

from dbdaora.hashring import HashRing

from . import GeoRadiusOutput, MemoryDataSource, MemoryMultiExec, RangeOutput


class AioRedisDataSource(Redis, MemoryDataSource):
    ...


class AioRedisMultiExec(MultiExec):
    ...


@dataclasses.dataclass
class ShardsAioRedisMultiExec(MemoryMultiExec):
    hashring: HashRing[AioRedisMultiExec]
    futures: List[Any] = dataclasses.field(default_factory=list)
    clients_to_execute: Set[AioRedisMultiExec] = dataclasses.field(
        default_factory=set
    )

    def get_client(self, key: str) -> AioRedisMultiExec:
        return self.hashring.get_node(key)

    def delete(self, key: str) -> Any:
        client = self.get_client(key)
        future = client.delete(key)
        self.clients_to_execute.add(client)
        self.futures.append(future)
        return future

    def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> Any:
        client = self.get_client(key)
        future = client.hmset(key, field, value, *pairs)
        self.clients_to_execute.add(client)
        self.futures.append(future)
        return future

    def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> Any:
        client = self.get_client(key)
        future = client.zadd(key, score, member, *pairs)
        self.clients_to_execute.add(client)
        self.futures.append(future)
        return future

    async def execute(self, *, return_exceptions: bool = False) -> Any:
        await asyncio.gather(
            *[
                client.execute(return_exceptions=return_exceptions)
                for client in self.clients_to_execute
            ]
        )
        self.clients_to_execute.clear()

        results = await asyncio.gather(*self.futures)
        self.futures.clear()
        return results


@dataclasses.dataclass
class ShardsAioRedisDataSource(MemoryDataSource):
    hashring: HashRing[AioRedisDataSource]
    geopoint_cls: ClassVar[Type[GeoPoint]] = GeoPoint  # type: ignore
    geomember_cls: ClassVar[Type[GeoMember]] = GeoMember  # type: ignore

    def get_client(self, key: str) -> AioRedisDataSource:
        return self.hashring.get_node(key)

    async def get(self, key: str) -> Optional[bytes]:
        return await self.get_client(key).get(key)

    async def set(self, key: str, data: str) -> None:
        await self.get_client(key).set(key, data)

    async def delete(self, key: str) -> None:
        await self.get_client(key).delete(key)

    async def expire(self, key: str, time: int) -> None:
        await self.get_client(key).expire(key, time)

    async def exists(self, key: str) -> int:
        return await self.get_client(key).exists(key)

    async def zrevrange(
        self, key: str, start: int, stop: int, withscores: bool = False
    ) -> Optional[RangeOutput]:
        return await self.get_client(key).zrevrange(
            key, start=start, stop=stop, withscores=withscores
        )

    async def zrange(
        self,
        key: str,
        start: int = 0,
        stop: int = -1,
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        return await self.get_client(key).zrange(
            key, start=start, stop=stop, withscores=withscores
        )

    async def zadd(
        self, key: str, score: float, member: str, *pairs: Union[float, str]
    ) -> None:
        await self.get_client(key).zadd(key, score, member, *pairs)

    async def zcard(self, key: str) -> int:
        return await self.get_client(key).zcard(key)

    async def zrevrangebyscore(
        self,
        key: str,
        max: float = float('inf'),
        min: float = float('-inf'),
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        return await self.get_client(key).zrevrangebyscore(
            key, max=max, min=min, withscores=withscores
        )

    async def zrangebyscore(
        self,
        key: str,
        min: float = float('-inf'),
        max: float = float('inf'),
        withscores: bool = False,
    ) -> Optional[RangeOutput]:
        return await self.get_client(key).zrangebyscore(
            key, min=min, max=max, withscores=withscores
        )

    async def hmset(
        self,
        key: str,
        field: Union[str, bytes],
        value: Union[str, bytes],
        *pairs: Union[str, bytes],
    ) -> None:
        await self.get_client(key).hmset(key, field, value, *pairs)

    async def hmget(
        self, key: str, field: Union[str, bytes], *fields: Union[str, bytes]
    ) -> Sequence[Optional[bytes]]:
        return await self.get_client(key).hmget(key, field, *fields)

    async def hgetall(self, key: str) -> Dict[bytes, bytes]:
        return await self.get_client(key).hgetall(key)

    async def georadius(
        self,
        key: str,
        longitude: float,
        latitude: float,
        radius: float,
        unit: str = 'm',
        *,
        with_dist: bool = False,
        with_coord: bool = False,
        count: Optional[int] = None,
    ) -> GeoRadiusOutput:
        return await self.get_client(key).georadius(
            key=key,
            longitude=longitude,
            latitude=latitude,
            radius=radius,
            unit=unit,
            with_dist=with_dist,
            with_coord=with_coord,
            count=count,
        )

    async def geoadd(
        self,
        key: str,
        longitude: float,
        latitude: float,
        member: Union[str, bytes],
        *args: Any,
        **kwargs: Any,
    ) -> int:
        return await self.get_client(key).geoadd(
            key, longitude, latitude, member, *args, **kwargs,
        )

    def close(self) -> None:
        for client in self.hashring.nodes:
            client.close()

    async def wait_closed(self) -> None:
        for client in self.hashring.nodes:
            await client.wait_closed()

    def multi_exec(self) -> MemoryMultiExec:
        hashring = type(self.hashring)(
            [
                AioRedisMultiExec(node._pool_or_conn, node.__class__)  # type: ignore
                for node in self.hashring.nodes
            ],
            self.hashring.nodes_size,
        )
        return ShardsAioRedisMultiExec(hashring)  # type: ignore


async def make(
    *uris: str,
    hashring_cls: Type[HashRing[AioRedisDataSource]] = HashRing,
    hashring_nodes_size: Optional[int] = None,
    commands_factory: Type[
        ShardsAioRedisDataSource
    ] = ShardsAioRedisDataSource,
) -> Union[Redis, ShardsAioRedisDataSource]:
    if len(uris) == 0:
        uris = ['redis://']  # type: ignore

    if len(uris) > 1:
        clients: Sequence[AioRedisDataSource] = [
            await create_redis_pool(  # type: ignore
                uri, commands_factory=AioRedisDataSource
            )
            for uri in uris
        ]
        hashring = hashring_cls(clients, hashring_nodes_size)
        return commands_factory(hashring)

    else:
        return await create_redis_pool(
            uris[0], commands_factory=AioRedisDataSource
        )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['dbdaora',
 'dbdaora.boolean',
 'dbdaora.boolean.repositories',
 'dbdaora.data_sources',
 'dbdaora.data_sources.fallback',
 'dbdaora.data_sources.memory',
 'dbdaora.geospatial',
 'dbdaora.geospatial.repositories',
 'dbdaora.geospatial.service',
 'dbdaora.hash',
 'dbdaora.hash.repositories',
 'dbdaora.hash.service',
 'dbdaora.service',
 'dbdaora.sorted_set',
 'dbdaora.sorted_set.repositories',
 'dbdaora.sorted_set.service']

package_data = \
{'': ['*'],
 'dbdaora': ['_tests/*'],
 'dbdaora.boolean': ['_tests/*', '_tests/datastore/*'],
 'dbdaora.boolean.repositories': ['_tests/*'],
 'dbdaora.geospatial': ['_tests/*', '_tests/datastore/*', '_tests/mongodb/*'],
 'dbdaora.geospatial.repositories': ['_tests/*'],
 'dbdaora.hash': ['_tests/*', '_tests/datastore/*', '_tests/mongodb/*'],
 'dbdaora.hash.repositories': ['_tests/*'],
 'dbdaora.sorted_set.repositories': ['_tests/*'],
 'dbdaora.sorted_set.service': ['_tests/*',
                                '_tests/datastore/*',
                                '_tests/mongodb/*']}

install_requires = \
['circuitbreaker', 'cachetools', 'jsondaora']

extras_require = \
{'aioredis': ['aioredis'],
 'datastore': ['google-cloud-datastore'],
 'doc': ['mkdocs', 'mkdocs-material', 'markdown-include'],
 'mongodb': ['motor'],
 'test': ['black',
          'isort',
          'ipython',
          'mypy',
          'pytest-asyncio',
          'pytest-cov',
          'pytest-mock',
          'pytest']}

setup(name='dbdaora',
      version='0.21.0',
      description='Communicates with databases using repository pattern and service patterns',
      author='Diogo Dutra',
      author_email='diogodutradamata@gmail.com',
      url='https://github.com/dutradda/dbdaora',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



from logging import Logger, getLogger
from typing import Any, Optional, Tuple, Type, Union

from cachetools import Cache

from ..cache import CacheType
from ..circuitbreaker import AsyncCircuitBreaker
from ..entity import Entity, EntityData
from ..keys import FallbackKey
from ..repository import MemoryRepository
from . import Service


async def build(
    service_cls: Type[Service[Entity, EntityData, FallbackKey]],
    repository_cls: Type[MemoryRepository[Entity, EntityData, FallbackKey]],
    memory_data_source_factory: Any,
    fallback_data_source_factory: Any,
    repository_expire_time: int,
    cache_type: Optional[CacheType] = None,
    cache_ttl: Optional[int] = None,
    cache_max_size: Optional[int] = None,
    cb_failure_threshold: Optional[int] = None,
    cb_recovery_timeout: Optional[int] = None,
    cb_expected_exception: Optional[
        Union[Type[Exception], Tuple[Type[Exception], ...]]
    ] = None,
    cb_expected_fallback_exception: Optional[
        Union[Type[Exception], Tuple[Type[Exception], ...]]
    ] = None,
    logger: Logger = getLogger(__name__),
    cache_ttl_failure_threshold: int = 0,
) -> Service[Entity, EntityData, FallbackKey]:
    repository = await build_repository(
        repository_cls,
        memory_data_source_factory,
        fallback_data_source_factory,
        repository_expire_time,
    )

    if cb_expected_exception and cb_expected_fallback_exception:
        if not isinstance(cb_expected_exception, tuple):
            cb_expected_exception = (cb_expected_exception,)

        if not isinstance(cb_expected_fallback_exception, tuple):
            cb_expected_fallback_exception = (cb_expected_fallback_exception,)

        cb_expected_exception += cb_expected_fallback_exception

    circuit_breaker = build_circuit_breaker(
        f'{repository_cls.name}_memory',
        cb_failure_threshold,
        cb_recovery_timeout,
        cb_expected_exception,
    )
    fallback_circuit_breaker = build_circuit_breaker(
        f'{repository_cls.name}_fallback',
        cb_failure_threshold,
        cb_recovery_timeout,
        cb_expected_fallback_exception,
    )
    cache = build_cache(
        cache_type, cache_ttl, cache_max_size, cache_ttl_failure_threshold
    )
    exists_cache = build_cache(
        cache_type, cache_ttl, cache_max_size, cache_ttl_failure_threshold
    )
    return service_cls(
        repository,
        circuit_breaker,
        fallback_circuit_breaker,
        cache,
        exists_cache,
        logger,
    )


async def build_repository(
    cls: Type[MemoryRepository[Entity, EntityData, FallbackKey]],
    memory_data_source_factory: Any,
    fallback_data_source_factory: Any,
    expire_time: int,
) -> MemoryRepository[Entity, EntityData, FallbackKey]:
    return cls(
        memory_data_source=await memory_data_source_factory(),
        fallback_data_source=await fallback_data_source_factory(),
        expire_time=expire_time,
    )


def build_cache(
    cache_type: Optional[CacheType] = None,
    ttl: Optional[int] = None,
    max_size: Optional[int] = None,
    ttl_failure_threshold: int = 0,
) -> Optional[Cache]:
    if cache_type:
        if max_size is None:
            raise Exception()

        if cache_type == CacheType.TTL or cache_type == CacheType.TTLDAORA:
            if ttl is None:
                raise Exception()

            if cache_type == CacheType.TTL:
                return cache_type.value(max_size, ttl)

            if cache_type == CacheType.TTLDAORA:
                return cache_type.value(max_size, ttl, ttl_failure_threshold)

        else:
            return cache_type.value(max_size)

    return None


def build_circuit_breaker(
    name: str,
    failure_threshold: Optional[int] = None,
    recovery_timeout: Optional[int] = None,
    expected_exception: Optional[
        Union[Type[Exception], Tuple[Type[Exception], ...]]
    ] = None,
) -> AsyncCircuitBreaker:
    return AsyncCircuitBreaker(
        failure_threshold, recovery_timeout, expected_exception, name=name,
    )



from typing import (  # type: ignore
    Any,
    ClassVar,
    Dict,
    Optional,
    Sequence,
    Type,
    Union,
    _TypedDictMeta,
)

from dbdaora.data_sources.memory import GeoMember
from dbdaora.exceptions import (
    EntityNotFoundError,
    InvalidGeoSpatialDataError,
    InvalidQueryError,
)
from dbdaora.keys import FallbackKey
from dbdaora.query import BaseQuery, Query
from dbdaora.repository import MemoryRepository

from ..entity import GeoSpatialData, GeoSpatialEntity


class GeoSpatialRepository(MemoryRepository[Any, GeoSpatialData, FallbackKey]):
    __skip_cls_validation__ = ('GeoSpatialRepository',)
    entity_cls: ClassVar[Type[Any]] = GeoSpatialEntity

    async def get_memory_data(  # type: ignore
        self, key: str, query: 'GeoSpatialQuery[FallbackKey]',
    ) -> Optional[GeoSpatialData]:
        if query.type == GeoSpatialQueryType.RADIUS:
            if (
                query.latitude is None
                or query.longitude is None
                or query.max_distance is None
            ):
                raise InvalidQueryError(query)

        else:
            raise InvalidQueryError(query)

        data = await self.memory_data_source.georadius(
            key=key,
            longitude=query.longitude,
            latitude=query.latitude,
            radius=query.max_distance,
            unit=query.distance_unit,
            with_dist=query.with_dist,
            with_coord=query.with_coord,
            count=query.count,
        )

        if not data:
            return None

        return data

    async def get_fallback_data(  # type: ignore
        self,
        query: 'GeoSpatialQuery[FallbackKey]',
        *,
        for_memory: bool = False,
    ) -> Optional[GeoSpatialData]:
        key = self.fallback_key(query)
        data = tuple(await self.fallback_data_source.query(key))

        if not data:
            return None

        return self.make_fallback_data_for_memory(key, query, data)

    def make_fallback_data_for_memory(
        self,
        key: FallbackKey,
        query: 'GeoSpatialQuery[FallbackKey]',
        data: Sequence[Dict[str, Any]],
    ) -> GeoSpatialData:
        return [
            self.memory_data_source.geomember_cls(
                member=member['member'],
                hash=None,
                dist=None,
                coord=self.memory_data_source.geopoint_cls(
                    latitude=member['latitude'], longitude=member['longitude'],
                ),
            )
            for member in data
        ]

    def make_entity(  # type: ignore
        self,
        data: GeoSpatialData,
        query: 'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
    ) -> Any:
        return self.entity_cls(
            data=data,
            **{
                id_name: id_value
                for id_name, id_value in zip(self.key_attrs, query.key_parts)
            },
        )

    def make_entity_from_fallback(  # type: ignore
        self,
        data: GeoSpatialData,
        query: 'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
    ) -> Any:
        return self.make_entity(data, query)

    async def add_memory_data(
        self, key: str, data: GeoSpatialData, from_fallback: bool = False
    ) -> None:
        if (
            isinstance(data, self.memory_data_source.geomember_cls)
            and data.coord is not None
        ):
            await self.memory_data_source.geoadd(
                key,
                longitude=data.coord.longitude,
                latitude=data.coord.latitude,
                member=data.member,
            )
        else:
            raise InvalidGeoSpatialDataError(data)

    async def add_memory_data_from_fallback(
        self,
        key: str,
        query: Union[
            BaseQuery[GeoSpatialEntity, GeoSpatialData, FallbackKey],
            GeoSpatialEntity,
        ],
        data: GeoSpatialData,
    ) -> GeoSpatialData:
        geomembers = self.make_memory_data_from_fallback(query, data)

        for i, geomember in enumerate(geomembers):
            if (
                isinstance(geomember, self.memory_data_source.geomember_cls)
                and geomember.coord is not None
            ):
                await self.memory_data_source.geoadd(
                    key,
                    longitude=geomember.coord.longitude,
                    latitude=geomember.coord.latitude,
                    member=geomember.member,
                )
            else:
                raise InvalidGeoSpatialDataError(i, geomember)

        if isinstance(query, GeoSpatialQuery):
            memory_data = await self.get_memory_data(key, query)
        else:
            memory_data = data

        if memory_data is None:
            raise EntityNotFoundError(query)

        return memory_data

    def make_memory_data_from_fallback(
        self,
        query: Union[
            BaseQuery[GeoSpatialEntity, GeoSpatialData, FallbackKey],
            GeoSpatialEntity,
        ],
        data: GeoSpatialData,
    ) -> Sequence[GeoMember]:
        return data  # type: ignore

    def make_memory_data_from_entity(
        self, entity: GeoSpatialEntity
    ) -> GeoSpatialData:
        return entity.data

    async def add_fallback(
        self,
        entity: GeoSpatialEntity,
        *entities: GeoSpatialEntity,
        **kwargs: Any,
    ) -> None:
        if (
            isinstance(entity.data, self.memory_data_source.geomember_cls)
            and entity.data.coord is not None
        ):
            data = {
                'latitude': entity.data.coord.latitude,
                'longitude': entity.data.coord.longitude,
                'member': entity.data.member,
            }
            await self.fallback_data_source.put(
                self.fallback_key(entity), data, **kwargs
            )
            return

        raise InvalidGeoSpatialDataError(entity)

    def fallback_key(
        self,
        query: Union[
            'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
            GeoSpatialEntity,
            Dict[str, Any],
        ],
    ) -> FallbackKey:
        if isinstance(query, Query):
            return self.fallback_data_source.make_key(
                self.name, *query.key_parts, *('',)
            )

        elif isinstance(self.get_entity_type(query), _TypedDictMeta):
            return self.fallback_data_source.make_key(
                self.name,
                *self.key_parts(query),
                *(
                    query['data'].member.decode()  # type: ignore
                    if isinstance(query['data'].member, bytes)  # type: ignore
                    else query['data'].member,  # type: ignore
                ),
            )

        elif isinstance(query, self.get_entity_type(query)):
            return self.fallback_data_source.make_key(
                self.name,
                *self.key_parts(query),
                *(
                    query.data.member.decode()
                    if isinstance(query.data.member, bytes)
                    else query.data.member,
                ),
            )

        raise InvalidQueryError(query)

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> 'BaseQuery[GeoSpatialEntity, GeoSpatialData, FallbackKey]':
        return query_factory(self, *args, **kwargs)

    async def already_got_not_found(
        self,
        query: Union[
            'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
            GeoSpatialEntity,
        ],
    ) -> bool:
        return bool(
            await self.memory_data_source.exists(self.memory_key(query))
        )

    async def delete_fallback_not_found(
        self,
        query: Union[
            'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
            GeoSpatialEntity,
        ],
    ) -> None:
        ...

    async def set_fallback_not_found(
        self,
        query: Union[
            'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
            GeoSpatialEntity,
        ],
    ) -> None:
        ...

    async def delete(
        self, query: 'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def exists(
        self, query: 'Query[GeoSpatialEntity, GeoSpatialData, FallbackKey]',
    ) -> bool:
        raise NotImplementedError()  # pragma: no cover


from ..query import (  # noqa isort:skip
    GeoSpatialQuery,
    GeoSpatialQueryType,
)
from ..query import make as query_factory  # noqa isort:skip



import pytest

from dbdaora import (
    DictFallbackDataSource,
    SortedSetEntity,
    SortedSetRepository,
)


@pytest.fixture
def fake_entity_cls():
    return SortedSetEntity


@pytest.fixture
def fake_entity(fake_entity_cls):
    return fake_entity_cls(id='fake', values=[b'1', b'2'])


@pytest.fixture
def fake_entity_withscores(fake_entity_cls):
    return fake_entity_cls(id='fake', values=[(b'1', 0), (b'2', 1)])


@pytest.fixture
def fake_repository_cls(fake_entity_cls):
    class FakeRepository(SortedSetRepository[str]):
        entity_cls = fake_entity_cls

    return FakeRepository


@pytest.fixture
def fallback_data_source():
    return DictFallbackDataSource()



from typing import Any, ClassVar, Dict, Generic, Iterable, Optional

from dbdaora.keys import FallbackKey

from .. import DataSource


class FallbackDataSource(DataSource, Generic[FallbackKey]):
    key_separator: ClassVar[str] = ':'

    def make_key(self, *key_parts: Any) -> FallbackKey:
        raise NotImplementedError()  # pragma: no cover

    async def get(self, key: FallbackKey) -> Optional[Dict[str, Any]]:
        raise NotImplementedError()  # pragma: no cover

    async def put(
        self, key: FallbackKey, data: Dict[str, Any], **kwargs: Any
    ) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def delete(self, key: FallbackKey) -> None:
        raise NotImplementedError()  # pragma: no cover

    async def query(
        self, key: FallbackKey, **kwargs: Any
    ) -> Iterable[Dict[str, Any]]:
        raise NotImplementedError()  # pragma: no cover



from typing import Any

from ...entity import Entity
from ...keys import FallbackKey
from ...service import Service
from ..repositories import HashData


class HashService(Service[Entity, HashData, FallbackKey]):
    def cache_key_suffix(self, **filters: Any) -> str:
        return (
            ''.join(f'{f}{v}' for f, v in filters.items() if f != 'fields')
            if filters
            else ''
        )

    def get_cached_entity(
        self, id: str, key_suffix: str, **filters: Any,
    ) -> Any:
        entity = super().get_cached_entity(id, key_suffix, **filters)
        fields = filters.get('fields')

        if fields is None:
            return entity

        if isinstance(entity, dict):
            for field in fields:
                if field not in entity:
                    return None

        else:
            for field in fields:
                if hasattr(entity, field):
                    return None

        return entity



from logging import Logger, getLogger
from typing import Any, Optional, Type

from dbdaora.entity import Entity, EntityData
from dbdaora.keys import FallbackKey
from dbdaora.service.builder import build as build_base_service

from ..cache import CacheType
from ..repository import MemoryRepository
from ..service import Service
from .service import HashService


async def make_service(
    repository_cls: Type[MemoryRepository[Entity, EntityData, FallbackKey]],
    memory_data_source_factory: Any,
    fallback_data_source_factory: Any,
    repository_expire_time: int,
    cache_type: Optional[CacheType] = None,
    cache_ttl: Optional[int] = None,
    cache_max_size: Optional[int] = None,
    cb_failure_threshold: Optional[int] = None,
    cb_recovery_timeout: Optional[int] = None,
    cb_expected_exception: Optional[Type[Exception]] = None,
    logger: Logger = getLogger(__name__),
) -> Service[Entity, EntityData, FallbackKey]:
    return await build_base_service(
        HashService,  # type: ignore
        repository_cls,
        memory_data_source_factory,
        fallback_data_source_factory,
        repository_expire_time,
        cache_type=cache_type,
        cache_ttl=cache_ttl,
        cache_max_size=cache_max_size,
        cb_failure_threshold=cb_failure_threshold,
        cb_recovery_timeout=cb_recovery_timeout,
        cb_expected_exception=cb_expected_exception,
        logger=logger,
    )



import dataclasses
from enum import Enum
from typing import Any, List, Optional

from dbdaora.keys import FallbackKey
from dbdaora.query import BaseQuery, Query


class GeoSpatialQueryType(Enum):
    RADIUS = 'radius'


@dataclasses.dataclass(init=False)
class GeoSpatialQuery(Query[Any, 'GeoSpatialData', FallbackKey]):
    repository: 'GeoSpatialRepository[FallbackKey]'
    type: GeoSpatialQueryType = GeoSpatialQueryType.RADIUS
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    max_distance: Optional[float] = None
    distance_unit: str = 'km'
    with_dist: bool = True
    with_coord: bool = True
    count: Optional[int] = None

    def __init__(
        self,
        repository: 'GeoSpatialRepository[FallbackKey]',
        *args: Any,
        memory: bool = True,
        key_parts: Optional[List[Any]] = None,
        type: GeoSpatialQueryType = GeoSpatialQueryType.RADIUS,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        max_distance: Optional[float] = None,
        distance_unit: str = 'km',
        with_dist: bool = True,
        with_coord: bool = True,
        count: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(
            repository, memory=memory, key_parts=key_parts, *args, **kwargs,
        )
        self.type = type
        self.latitude = latitude
        self.longitude = longitude
        self.max_distance = max_distance
        self.distance_unit = distance_unit
        self.with_dist = with_dist
        self.with_coord = with_coord
        self.count = count


def make(
    *args: Any, **kwargs: Any
) -> BaseQuery[Any, 'GeoSpatialData', FallbackKey]:
    return GeoSpatialQuery(*args, **kwargs)


from .repositories import GeoSpatialRepository  # noqa isort:skip
from .entity import GeoSpatialData  # noqa isort:skip



class DBDaoraError(Exception):
    ...


class EntityNotFoundError(DBDaoraError):
    ...


class InvalidQueryError(DBDaoraError):
    ...


class InvalidHashAttribute(DBDaoraError):
    ...


class InvalidEntityAnnotationError(DBDaoraError):
    ...


class RequiredKeyAttributeError(DBDaoraError):
    ...


class InvalidKeyAttributeError(DBDaoraError):
    ...


class InvalidEntityTypeError(DBDaoraError):
    ...


class RequiredClassAttributeError(DBDaoraError):
    ...


class InvalidGeoSpatialDataError(DBDaoraError):
    ...


class CacheNotAvailableError(DBDaoraError):
    ...



import dataclasses
from typing import Optional, Sequence, TypedDict, Union

from jsondaora import jsondaora

from dbdaora.data_sources.memory import RangeOutput


SortedSetInput = Sequence[Union[str, float]]

SortedSetData = Union[RangeOutput, SortedSetInput]


@dataclasses.dataclass
class SortedSetEntity:
    id: str
    values: SortedSetData
    max_size: Optional[int] = None


@jsondaora
class SortedSetDictEntity(TypedDict):
    id: str
    values: SortedSetData
    max_size: Optional[int]


Entity = Union[SortedSetEntity, SortedSetDictEntity]



from typing import Any, AsyncGenerator, Optional, Union

from dbdaora import EntityNotFoundError
from dbdaora.keys import FallbackKey
from dbdaora.query import BaseQuery, Query, QueryMany
from dbdaora.query import make as query_factory
from dbdaora.repository import MemoryRepository


class BooleanRepository(MemoryRepository[Any, bool, FallbackKey]):
    __skip_cls_validation__ = ('BooleanRepository',)

    async def get_memory_data(
        self, key: str, query: BaseQuery[Any, bool, FallbackKey],
    ) -> Optional[bool]:
        value = await self.memory_data_source.get(key)
        return None if value is None else bool(int(value))

    async def get_fallback_data(
        self,
        query: BaseQuery[Any, bool, FallbackKey],
        *,
        for_memory: bool = False,
    ) -> Optional[bool]:
        data = await self.fallback_data_source.get(self.fallback_key(query))

        if data is not None:
            return True

        return None

    def make_entity(  # type: ignore
        self, data: bool, query: Query[Any, bool, FallbackKey],
    ) -> Any:
        return query.key_parts[-1]

    def make_entity_from_fallback(  # type: ignore
        self, data: bool, query: Query[Any, bool, FallbackKey],
    ) -> Any:
        return self.make_entity(data, query)

    async def add_memory_data(
        self, key: str, data: bool, from_fallback: bool = False
    ) -> None:
        await self.memory_data_source.set(key, '1')

    async def add_memory_data_from_fallback(
        self,
        key: str,
        query: Union[BaseQuery[Any, bool, FallbackKey], Any],
        data: bool,
    ) -> bool:
        await self.add_memory_data(key, data)
        return True

    def make_memory_data_from_entity(self, entity: Any) -> bool:
        return True

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await self.fallback_data_source.put(
            self.fallback_key(entity), {'value': True}, **kwargs
        )

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> BaseQuery[Any, bool, FallbackKey]:
        return query_factory(self, *args, **kwargs)

    async def add_memory(
        self, entity: Any, *entities: Any, fallback_ttl: Optional[int] = None
    ) -> None:
        memory_key = self.memory_key(entity)
        memory_data = self.make_memory_data_from_entity(entity)

        await self.add_memory_data(memory_key, memory_data)
        await self.set_expire_time(memory_key)
        await self.add_fallback(entity, fallback_ttl=fallback_ttl)

    async def set_fallback_not_found(
        self, query: Union[BaseQuery[Any, bool, FallbackKey], Any],
    ) -> None:
        memory_key = self.memory_key(query)
        await self.memory_data_source.set(memory_key, '0')
        await self.set_expire_time(memory_key)

    async def get_memory_many(
        self, query: QueryMany[Any, bool, FallbackKey],
    ) -> AsyncGenerator[Any, None]:
        for query_i in query.queries:
            memory_key = self.memory_key(query_i)
            memory_data = await self.get_memory_data(memory_key, query_i)

            if memory_data:
                yield self.make_entity(memory_data, query_i)

            else:
                fallback_data = await self.get_fallback_data(
                    query_i, for_memory=True
                )

                if fallback_data is None:
                    await self.set_fallback_not_found(query_i)
                else:
                    memory_data = await self.add_memory_data_from_fallback(
                        memory_key, query_i, fallback_data
                    )
                    await self.set_expire_time(memory_key)
                    yield self.make_entity(memory_data, query_i)

    async def get_memory(self, query: Query[Any, bool, FallbackKey]) -> Any:
        memory_key = self.memory_key(query)
        memory_data = await self.get_memory_data(memory_key, query)

        if memory_data is None:
            fallback_data = await self.get_fallback_data(
                query, for_memory=True
            )

            if fallback_data is None:
                await self.set_fallback_not_found(query)
            else:
                memory_data = await self.add_memory_data_from_fallback(
                    memory_key, query, fallback_data
                )
                await self.set_expire_time(memory_key)

        if not memory_data:
            raise EntityNotFoundError(query)

        return self.make_entity(memory_data, query)

    async def delete(self, query: Query[Any, bool, FallbackKey],) -> None:
        if query.memory:
            await self.set_fallback_not_found(query)

        await self.fallback_data_source.delete(self.fallback_key(query))

    def fallback_not_found_key(
        self, query: Union[BaseQuery[Any, bool, FallbackKey], Any],
    ) -> str:
        return self.memory_key(query)

    async def delete_fallback_not_found(
        self, query: Union[BaseQuery[Any, bool, FallbackKey], Any],
    ) -> None:
        ...



import asyncio
import itertools
from typing import (
    Any,
    Awaitable,
    ClassVar,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    Union,
)

from dbdaora.keys import FallbackKey
from dbdaora.repository import MemoryRepository

from ..entity import SortedSetData, SortedSetEntity
from ..query import SortedSetQuery


class FallbackSortedSetData(TypedDict):
    values: Sequence[Union[str, float]]


class SortedSetRepository(MemoryRepository[Any, SortedSetData, FallbackKey]):
    __skip_cls_validation__ = ('SortedSetRepository',)
    entity_cls: ClassVar[Type[Any]] = SortedSetEntity

    async def get_memory_data(  # type: ignore
        self, key: str, query: SortedSetQuery[FallbackKey],
    ) -> Optional[SortedSetData]:
        size_task: Optional[asyncio.Task[Any]] = None
        data_task: Awaitable[Any]

        if query.withmaxsize:
            size_task = asyncio.create_task(self.memory_data_source.zcard(key))

        if query.max_score is not None or query.min_score is not None:
            max_score, min_score = self.parse_score_limits(query)

            if query.reverse:
                data_task = asyncio.create_task(
                    self.memory_data_source.zrevrangebyscore(
                        key,
                        max=max_score,
                        min=min_score,
                        withscores=query.withscores,
                    )
                )
            else:
                data_task = asyncio.create_task(
                    self.memory_data_source.zrangebyscore(
                        key,
                        max=max_score,
                        min=min_score,
                        withscores=query.withscores,
                    )
                )

        else:
            start, stop = self.parse_page(query)

            if query.reverse:
                data_task = asyncio.create_task(
                    self.memory_data_source.zrevrange(
                        key,
                        start=start,
                        stop=stop,
                        withscores=query.withscores,
                    )
                )
            else:
                data_task = asyncio.create_task(
                    self.memory_data_source.zrange(
                        key,
                        start=start,
                        stop=stop,
                        withscores=query.withscores,
                    )
                )

        data = await data_task

        if not data:
            if size_task:
                size_task.cancel()
            return None

        return data, await size_task if size_task else None

    def parse_page(
        self, query: SortedSetQuery[FallbackKey]
    ) -> Tuple[int, int]:
        if query.page_size:
            if query.page == 1 or query.page is None:
                return 0, query.page_size - 1
            else:
                return (
                    (query.page - 1) * query.page_size,
                    query.page * query.page_size - 1,
                )

        return 0, -1

    def parse_score_limits(
        self, query: SortedSetQuery[FallbackKey]
    ) -> Tuple[float, float]:
        return (
            float('inf') if query.max_score is None else query.max_score,
            float('-inf') if query.min_score is None else query.min_score,
        )

    async def get_fallback_data(
        self,
        query: Union[SortedSetQuery[FallbackKey], Any],
        for_memory: bool = False,
    ) -> Optional[SortedSetData]:
        data: Optional[FallbackSortedSetData]

        data = await self.fallback_data_source.get(  # type: ignore
            self.fallback_key(query)
        )

        if data is None:
            return None

        data_withscores = [
            (
                data['values'][i].encode()  # type: ignore
                if isinstance(data['values'][i], str)
                else data['values'][i],
                data['values'][i + 1],
            )
            for i in range(0, len(data['values']), 2)
        ]

        if for_memory:
            return data_withscores  # type: ignore

        return self.parse_data_from_fallback(data_withscores, query)

    def parse_data_from_fallback(
        self, data_withscores: Any, query: Any
    ) -> Optional[SortedSetData]:
        sorted_data = sorted(
            data_withscores, key=lambda v: v[1], reverse=query.reverse
        )
        maxsize = len(sorted_data) if query.withmaxsize else None

        if query.max_score is not None or query.min_score is not None:
            max_score, min_score = self.parse_score_limits(query)

            if max_score != float('inf') or min_score != float('-inf'):
                sorted_data = [
                    (member, score)
                    for member, score in sorted_data
                    if min_score <= score <= max_score
                ]

        else:
            start, stop = self.parse_page(query)

            if start != 0 or stop != -1:
                sorted_data = sorted_data[start : stop + 1]  # noqa

        if not sorted_data:
            return None

        if query.withscores:
            return (sorted_data, maxsize)  # type: ignore

        return ([member for member, score in sorted_data], maxsize)  # type: ignore

    def make_entity(  # type: ignore
        self, data: SortedSetData, query: SortedSetQuery[FallbackKey]
    ) -> Any:
        return self.get_entity_type(query)(
            values=data[0],
            max_size=data[1],
            **{
                id_name: id_value
                for id_name, id_value in zip(self.key_attrs, query.key_parts)
            },
        )

    def make_entity_from_fallback(  # type: ignore
        self, data: SortedSetData, query: SortedSetQuery[FallbackKey]
    ) -> Any:
        return self.make_entity(data, query)

    async def add_memory_data(self, key: str, data: SortedSetData) -> None:
        delete_task = asyncio.create_task(self.memory_data_source.delete(key))
        delete_task.add_done_callback(task_done_callback)
        zadd_task = asyncio.create_task(
            self.memory_data_source.zadd(key, *data)
        )
        zadd_task.add_done_callback(task_done_callback)
        await delete_task
        await zadd_task

    async def add_fallback(
        self, entity: Any, *entities: Any, **kwargs: Any
    ) -> None:
        await self.fallback_data_source.put(
            self.fallback_key(entity),
            {
                'values': list(itertools.chain(*entity['values']))
                if isinstance(entity, dict)
                else list(itertools.chain(*entity.values))
            },
            **kwargs,
        )

    async def add_memory_data_from_fallback(  # type: ignore
        self,
        key: str,
        query: Union[SortedSetQuery[FallbackKey], Any],
        data: Sequence[Tuple[str, float]],
    ) -> Optional[SortedSetData]:
        await self.memory_data_source.zadd(key, *self.format_memory_data(data))
        return self.parse_data_from_fallback(data, query)

    def make_query(
        self, *args: Any, **kwargs: Any
    ) -> SortedSetQuery[FallbackKey]:
        return SortedSetQuery(self, *args, **kwargs)

    def make_memory_data_from_entity(self, entity: Any) -> SortedSetData:
        if isinstance(entity, dict):
            return self.format_memory_data(entity['values'])

        return self.format_memory_data(entity.values)

    def format_memory_data(self, data: SortedSetData) -> SortedSetData:
        data = list(itertools.chain(*data))
        data.reverse()
        return data


def task_done_callback(f: Any) -> None:
    try:
        f.result()
    except Exception:  # pragma: no cover
        ...  # pragma: no cover
